unit Color4F;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls;

type
    TPointArray = array [1..1000000] of TPoint;   // Array of points.
    PPointArray = ^TPointArray;                   // Pointer to array of points.

    // Map region class.
    TRegion = class
        public
            num_points   : Integer;     // # points on the region's border.
            points       : PPointArray; // The points on the region's border.
            neighbors    : TList;       // List of neighboring TRegions.
            number       : Integer;     // Region number starting with 0.
            color_number : Integer;     // Color number for the region.

            constructor Create; virtual;
            function SnapTo(var X, Y : Integer) : Boolean;
            procedure AddPoint(X, Y : Integer);
            procedure FinishRegion;
            procedure DrawRegion(cvs : TCanvas; colors : array of TColor; show_region_numbers : Boolean);
            procedure LoadRegion(var in_file : TextFile);
            procedure SaveRegion(var out_file : TextFile);
            function HasLine(a1, b1, a2, b2 : Integer) : Boolean;
            procedure CheckNeighbor(rgn : TRegion);
            function ListNeighbors : String;
            procedure ShowNodeNumber(cvs : TCanvas);
    end;

  TMapForm = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    mnuFileNew: TMenuItem;
    mnuFileOpen: TMenuItem;
    mnuFileSaveAs: TMenuItem;
    N1: TMenuItem;
    mnuFileExit: TMenuItem;
    Color1: TMenuItem;
    mnuColorNodes: TMenuItem;
    dlgOpen: TOpenDialog;
    dlgSaveAs: TSaveDialog;
    Debug1: TMenuItem;
    mnuShowRegionNumbers: TMenuItem;
    mnuShowNeighborLists: TMenuItem;
    procedure mnuFileExitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SnapTo(var X, Y : Integer);
    function AddRegion : TRegion;
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormPaint(Sender: TObject);
    procedure LoadMap(file_name : String);
    procedure mnuFileOpenClick(Sender: TObject);
    procedure FindNeighbors;
    function AssignOneColor(rgn_i : Integer) : Boolean;
    procedure AssignColors;
    procedure mnuColorNodesClick(Sender: TObject);
    procedure mnuFileSaveAsClick(Sender: TObject);
    procedure SaveMap(file_name : String);
    procedure mnuFileNewClick(Sender: TObject);
    procedure ShowNeighborLists;
    procedure ShowNodeNumbers;
    procedure mnuShowRegionNumbersClick(Sender: TObject);
    procedure mnuShowNeighborListsClick(Sender: TObject);

  private
        last_x, last_y : Integer; // Last point when drawing.
        cur_x, cur_y   : Integer; // Current point when drawing.
        drawing_region : TRegion; // Region we are drawing.
        regions        : TList;   // List of regions.
        colors         : array [0..5] of TColor; // Colors to fill regions.

  public
    { Public declarations }
  end;

var
  MapForm: TMapForm;

implementation
{$R *.DFM}

// ********************
// * Map region class *
// ********************
// Set some defaults.
constructor TRegion.Create;
begin
    inherited Create;
    num_points   := 0;
    neighbors    := TList.Create;
    number       := -1;
    color_number := 0;
end;

// If this region contains a point within SNAP pixels in
// both the X and Y directions, change X and Y to its
// coordinates and return True.
function TRegion.SnapTo(var X, Y : Integer) : Boolean;
const
    SNAP = 5;
var
    pt : Integer;
begin
    for pt := 1 to num_points do
    begin
        if ((Abs(X - points[pt].X) < SNAP) and
            (Abs(Y - points[pt].Y) < SNAP)) then
        begin
            X := points[pt].X;
            Y := points[pt].Y;
            result := True;
            exit;
        end;
    end;
    result := False;
end;

// Add the point to the points array.
procedure TRegion.AddPoint(X, Y : Integer);
var
    new_points : PPointArray;
    pt         : Integer;
begin
    // Create a new array to hold the values.
    GetMem(new_points, (num_points + 1) * SizeOf(TPoint));

    // Copy the values into the new array.
    for pt := 1 to num_points do
        new_points^[pt] := points^[pt];

    // Add the new point.
    new_points^[num_points + 1].X := X;
    new_points^[num_points + 1].Y := Y;

    // Free the previous list.
    if (num_points > 0) then FreeMem(points);

    // Make points point to the new array;
    points := new_points;

    // Update num_points.
    num_points := num_points + 1;
end;

// Copy the first point into the last position to close
// the polygonal region.
procedure TRegion.FinishRegion;
begin
    AddPoint(points[1].X, points[1].Y);
end;

// Draw the region.
procedure TRegion.DrawRegion(cvs : TCanvas; colors : array of TColor; show_region_numbers : Boolean);
begin
    // Do nothing if we have no points.
    if (num_points < 2) then exit;

    // Fill the region with its color.
    cvs.Brush.Color := colors[color_number];
    cvs.Polygon(Slice(points^, num_points));

    // Display the region number.
    if (show_region_numbers) then ShowNodeNumber(cvs);
end;

// Read this region's data from the file.
procedure TRegion.LoadRegion(var in_file : TextFile);
var
    num_pts, pt, X, Y : Integer;
begin
    // Read the number of points in this region.
    Readln(in_file, num_pts);
    num_points := 0;

    // Read the points.
    for pt := 1 to num_pts do
    begin
        Readln(in_file, X, Y);
        AddPoint(X, Y);
    end;
end;

// Write this region's data into the file.
procedure TRegion.SaveRegion(var out_file : TextFile);
var
    pt : Integer;
begin
    // Save the number of points in this region.
    Writeln(out_file, num_points);

    // Save the points.
    for pt := 1 to num_points do
        Writeln(out_file, points[pt].X, ' ', points[pt].Y);
end;

// Return True if the region contains this line segment.
function TRegion.HasLine(a1, b1, a2, b2 : Integer) : Boolean;
var
    i, x1, y1, x2, y2 : Integer;
begin
    // Assume we will find it.
    Result := True;

    x1 := points^[1].X;
    y1 := points^[1].Y;
    for i := 2 to num_points do
    begin
        x2 := points^[i].X;
        y2 := points^[i].Y;
        if ((x1 = a1) and (y1 = b1) and
            (x2 = a2) and (y2 = b2)) then exit;
        if ((x2 = a1) and (y2 = b1) and
            (x1 = a2) and (y1 = b2)) then exit;

        x1 := x2;
        y1 := y2;
    end;

    // We didn't find the line.
    Result := False;
end;

// See if this region is adjacent to the indicated one.
// If so, add the regions to each other's adjacency lists.
procedure TRegion.CheckNeighbor(rgn : TRegion);
var
    i, x1, y1, x2, y2 : Integer;
begin
    x1 := points^[1].X;
    y1 := points^[1].Y;
    for i := 2 to num_points do
    begin
        x2 := points^[i].X;
        y2 := points^[i].Y;
        if (rgn.HasLine(x1, y1, x2, y2)) then
        begin
            // They are neighbors. Add them to each other's
            // adjacency lists.
            neighbors.Add(rgn);
            rgn.neighbors.Add(Self);
            exit;
        end;
        x1 := x2;
        y1 := y2;
    end;
end;

// Display a message listing this region's neighbors.
function TRegion.ListNeighbors : String;
var
    i   : Integer;
    nbr : TRegion;
begin
    result := Format('Region %d: ', [number]);
    for i := 0 to neighbors.Count - 1 do
    begin
        nbr := neighbors.Items[i];
        result := result + Format('%d ', [nbr.number]);
    end;
    result := result + #13#10;
end;

// Display the region's number.
procedure TRegion.ShowNodeNumber(cvs : TCanvas);
var
    i, X, Y : Integer;
    txt     : String;
begin
    X := 0;
    Y := 0;
    // Find average region coordinate.
    for i := 1 to num_points do
    begin
        X := X + points[i].X;
        Y := Y + points[i].Y;
    end;

    txt := Format('%d', [number]);
    X := Round(X / num_points - cvs.TextWidth(txt) / 2);
    Y := Round(Y / num_points - cvs.TextHeight(txt) / 2);
    cvs.TextOut(X, Y, txt);
end;

// ******************
// * TMapForm class *
// ******************
// Exit.
procedure TMapForm.mnuFileExitClick(Sender: TObject);
begin
    Close;
end;

procedure TMapForm.FormCreate(Sender: TObject);
begin
    // Draw in black.
    Canvas.Pen.Color := clRed;

    // Initialize the colors.
    colors[0] := clWhite;   // Color for unassigned regions.
    colors[1] := clLime;
    colors[2] := clBlue;
    colors[3] := clYellow;
    colors[4] := clFuchsia;
    colors[5] := clMaroon;

    // Start with an empty regions list.
    regions := TList.Create;
end;

// If there is an existing point within SNAP pixels in
// both the X and Y directions, change X and Y to its
// coordinates.
procedure TMapForm.SnapTo(var X, Y : Integer);
var
    i   : Integer;
    rgn : TRegion;
begin
    for i := 0 to regions.Count - 1 do
    begin
         rgn := regions.Items[i];
        if (rgn.SnapTo(X, Y)) then exit;
    end;
end;

// Create a new region, add it to the regions list,
// and set its number value. Return the new region.
function TMapForm.AddRegion : TRegion;
begin
    // Create the new region.
    result := TRegion.Create;
    result.number := regions.Count;
    regions.Add(result);
end;

// Start selecting a region's point.
procedure TMapForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
    SnapTo(X, Y);

    // If we are currently drawing a region and the user
    // clicks the right button, finish the region.
    if ((Button = mbRight) and
        (drawing_region <> nil)) then
    begin
        // Finish the region we're drawing.
        // Erase the last rubberband line.
        Canvas.Pen.Mode := pmXor;
        Canvas.MoveTo(last_x, last_y);
        Canvas.LineTo(cur_x, cur_y);
        Canvas.Pen.Mode := pmCopy;

        // Finish the current drawing region.
        drawing_region.FinishRegion;

        // Draw the closing segment.
        Canvas.MoveTo(last_x, last_y);
        cur_x := drawing_region.points^[
            drawing_region.num_points].X;
        cur_y := drawing_region.points^[
            drawing_region.num_points].Y;
        Canvas.LineTo(cur_x, cur_y);

        drawing_region := nil;
        exit;
    end;

    // If we have not started a region, start one.
    if (drawing_region = nil) then
    begin
        // Create a new region.
        drawing_region := AddRegion;
    end else begin
        // Redraw the last line in black.
        Canvas.MoveTo(last_x, last_y);
        Canvas.LineTo(cur_x, cur_y);
    end;

    // Add the new point to the current region.
    drawing_region.AddPoint(X, Y);

    // Record the next point's starting position.
    last_x := X;
    last_y := Y;
    cur_x := X;
    cur_y := Y;
end;

// If we are currently drawing, draw a rubberband line to
// show where the next point will be.
procedure TMapForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
    // Do nothing if we are not drawing.
    if (drawing_region = nil) then exit;

    // Erase the previous rubberband line.
    Canvas.Pen.Mode := pmXor;
    Canvas.MoveTo(last_x, last_y);
    Canvas.LineTo(cur_x, cur_y);

    // Record the new point's position.
    SnapTo(X, Y);
    cur_x := X;
    cur_y := Y;

    // Draw the new rubberband line.
    Canvas.MoveTo(last_x, last_y);
    Canvas.LineTo(cur_x, cur_y);
    Canvas.Pen.Mode := pmCopy;
end;

// Redraw the map.
procedure TMapForm.FormPaint(Sender: TObject);
var
    i                   : Integer;
    rgn                 : TRegion;
    show_region_numbers : Boolean;
begin
    show_region_numbers := mnuShowRegionNumbers.Checked;
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        rgn.DrawRegion(Canvas, colors, show_region_numbers);
    end;
end;

// Load a map file.
procedure TMapForm.LoadMap(file_name : String);
var
    in_file          : TextFile;
    num_regions, rgn : Integer;
    new_rgn          : TRegion;
begin
    AssignFile(in_file, file_name);
    Reset(in_file);

    // See how many regions there are.
    Readln(in_file, num_regions);

    // Read the regions.
    regions.Clear;
    for rgn := 1 to num_regions do
    begin
        new_rgn := AddRegion;

        // Make the new region read itself from the file.
        new_rgn.LoadRegion(in_file);
    end;

    // Close the map file.
    CloseFile(in_file);
    drawing_region := nil;
end;

// Load a map file and draw it.
procedure TMapForm.mnuFileOpenClick(Sender: TObject);
begin
    if (dlgOpen.Execute) then
    begin
        LoadMap(dlgOpen.FileName);
        Caption := 'Color4 [' +
            ExtractFileName(dlgOpen.FileName) + ']';
        Refresh;
    end;
end;

// Create the adjacency lists for the regions.
procedure TMapForm.FindNeighbors;
var
    i1, i2     : Integer;
    rgn1, rgn2 : TRegion;
begin
    // Compare each region to every other and see if they
    // are adjacent.
    for i1 := 0 to regions.Count - 2 do
    begin
        rgn1 := regions.Items[i1];
        for i2 := i1 + 1 to regions.Count - 1 do
        begin
            rgn2 := regions.Items[i2];
            rgn1.CheckNeighbor(rgn2);
        end;
    end;

    // Display the neighbor lists if desired.
    if (mnuShowNeighborLists.Checked) then
        ShowNeighborLists;
end;

// Assign a color for the node with the indicated index.
// Recursively assign colors for the other nodes. Return
// True if we find a valid assignment.
function TMapForm.AssignOneColor(rgn_i : Integer) : Boolean;
var
    color_num, nbr_i : Integer;
    rgn, nbr         : TRegion;
begin
    // If rgn_i >= regions.Count, then all regions have
    // been assigned and we have a valid solution.
    if (rgn_i >= regions.Count) then
    begin
        result := True;
        exit;
    end;

    // Try each possible color for this region.
    rgn := regions.Items[rgn_i];
    for color_num := 1 to 4 do
    begin
        // See if this color is available for this region.
        for nbr_i := 0 to rgn.neighbors.Count - 1 do
        begin
            // See if this neighbor has used the color already.
            nbr := rgn.neighbors.Items[nbr_i];
            if (nbr.color_number = color_num) then break;
        end;

        // See if the color is usable.
        if (nbr_i >= rgn.neighbors.Count) then
        begin
            // Assign this color to the region.
            rgn.color_number := color_num;

            // Recursively assign colors to the other regions.
            if (AssignOneColor(rgn_i + 1)) then
            begin
                // We found a valid assignment.
                result := True;
                exit;
            end;
        end;

        // If everything worked and we found a complete
        // assignment, we have already exited. Otherwise
        // we continue trying other colors.
    end;

    // Blank our color so we can try again later.
    rgn.color_number := 0;

    // We found no valid assignments.
    result := False;
end;

// Four-color the map.
procedure TMapForm.AssignColors;
var
    rgn_i : Integer;
    rgn   : TRegion;
begin
    // Reset all the colors and clear the neighbor lists.
    for rgn_i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[rgn_i];
        rgn.neighbors.Clear;
        rgn.color_number := 0;
    end;

    // Make the adjacency lists.
    FindNeighbors;

    // Only continue if there are regions.
    if (regions.Count > 0) then
    begin
        // Assign the first region a color.
        rgn := regions.Items[0];
        rgn.color_number := 1;

        // Assign the other colors using an
        // exhaustive search.
        if (not AssignOneColor(1)) then
            ShowMessage('Error: Could not find a valid coloring.');
    end;
end;

// Call AssignColors to five-color the map.
procedure TMapForm.mnuColorNodesClick(Sender: TObject);
begin
    Cursor := crHourglass;

    // Assign the colors and redraw the map.
    AssignColors;
    Refresh;

    Cursor := crDefault;
end;

// Save the map.
procedure TMapForm.mnuFileSaveAsClick(Sender: TObject);
begin
    if (dlgSaveAs.Execute) then
        SaveMap(dlgSaveAs.FileName);
end;

// Load a map file.
procedure TMapForm.SaveMap(file_name : String);
var
    out_file : TextFile;
    i        : Integer;
    rgn      : TRegion;
begin
    AssignFile(out_file, file_name);
    Rewrite(out_file);

    // Save the number of regions.
    Writeln(out_file, regions.Count);

    // Save the regions.
    for i := 0 to regions.Count - 1 do
    begin
        // Make the new region write itself into the file.
        rgn := regions.Items[i];
        rgn.SaveRegion(out_file);
    end;

    // Close the map file.
    CloseFile(out_file);
end;

// Start a fresh map.
procedure TMapForm.mnuFileNewClick(Sender: TObject);
begin
    regions.Clear;
    drawing_region := nil;
    Refresh;
end;

procedure TMapForm.ShowNeighborLists;
var
    rgn : TRegion;
    i   : Integer;
    txt : String;
begin
    txt := '';
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        txt := txt + rgn.ListNeighbors;
    end;
    ShowMessage(txt);
end;

// Show the node numbers.
procedure TMapForm.ShowNodeNumbers;
var
    rgn : TRegion;
    i   : Integer;
begin
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        rgn.ShowNodeNumber(Canvas);
    end;
end;

// Toggle the Show Region Numbers option.
procedure TMapForm.mnuShowRegionNumbersClick(Sender: TObject);
begin
    mnuShowRegionNumbers.Checked :=
        (not mnuShowRegionNumbers.Checked);
    Refresh;
end;

// Toggle the Show Neighbor Lists option.
procedure TMapForm.mnuShowNeighborListsClick(Sender: TObject);
begin
    mnuShowNeighborLists.Checked :=
        (not mnuShowNeighborLists.Checked);
end;

end.
