unit Color5F;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls;

type
    TPointArray = array [1..1000000] of TPoint;   // Array of points.
    PPointArray = ^TPointArray;                   // Pointer to array of points.

    // Map region class.
    TRegion = class
        public
            num_points   : Integer;     // # points on the region's border.
            points       : PPointArray; // The points on the region's border.
            neighbors    : TList;       // List of neighboring TRegions.
            number       : Integer;     // Region number starting with 0.
            color_number : Integer;     // Color number for the region.

            constructor Create; virtual;
            function SnapTo(var X, Y : Integer) : Boolean;
            procedure AddPoint(X, Y : Integer);
            procedure FinishRegion;
            procedure DrawRegion(cvs : TCanvas; colors : array of TColor; show_region_numbers : Boolean);
            procedure LoadRegion(var in_file : TextFile);
            procedure SaveRegion(var out_file : TextFile);
            function HasLine(a1, b1, a2, b2 : Integer) : Boolean;
            procedure CheckNeighbor(rgn : TRegion);
            procedure RemoveNeighbor(target : TRegion);
            procedure RemoveFromGraph;
            function FindNonAdjacents(var n1, n2 : TRegion) : Boolean;
            procedure AddNeighbor(nbr : TRegion);
            procedure AssociateWith(target : TRegion);
            function ListNeighbors : String;
            procedure ShowNodeNumber(cvs : TCanvas);
    end;

  TMapForm = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    mnuFileNew: TMenuItem;
    mnuFileOpen: TMenuItem;
    mnuFileSaveAs: TMenuItem;
    N1: TMenuItem;
    mnuFileExit: TMenuItem;
    Color1: TMenuItem;
    mnuColorNodes: TMenuItem;
    dlgOpen: TOpenDialog;
    dlgSaveAs: TSaveDialog;
    Debug1: TMenuItem;
    mnuShowRegionNumbers: TMenuItem;
    mnuShowNeighborLists: TMenuItem;
    procedure mnuFileExitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SnapTo(var X, Y : Integer);
    function AddRegion : TRegion;
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormPaint(Sender: TObject);
    procedure LoadMap(file_name : String);
    procedure mnuFileOpenClick(Sender: TObject);
    procedure FindNeighbors;
    procedure AssignColors;
    procedure mnuColorNodesClick(Sender: TObject);
    procedure mnuFileSaveAsClick(Sender: TObject);
    procedure SaveMap(file_name : String);
    procedure mnuFileNewClick(Sender: TObject);
    procedure ShowNeighborLists;
    procedure ShowNodeNumbers;
    procedure mnuShowRegionNumbersClick(Sender: TObject);
    procedure mnuShowNeighborListsClick(Sender: TObject);

  private
        last_x, last_y : Integer; // Last point when drawing.
        cur_x, cur_y   : Integer; // Current point when drawing.
        drawing_region : TRegion; // Region we are drawing.
        regions        : TList;   // List of regions.
        colors         : array [0..5] of TColor; // Colors to fill regions.

  public
    { Public declarations }
  end;

var
  MapForm: TMapForm;

implementation
{$R *.DFM}

// ********************
// * Map region class *
// ********************
// Set some defaults.
constructor TRegion.Create;
begin
    inherited Create;
    num_points   := 0;
    neighbors    := TList.Create;
    number       := -1;
    color_number := 0;
end;

// If this region contains a point within SNAP pixels in
// both the X and Y directions, change X and Y to its
// coordinates and return True.
function TRegion.SnapTo(var X, Y : Integer) : Boolean;
const
    SNAP = 5;
var
    pt : Integer;
begin
    for pt := 1 to num_points do
    begin
        if ((Abs(X - points[pt].X) < SNAP) and
            (Abs(Y - points[pt].Y) < SNAP)) then
        begin
            X := points[pt].X;
            Y := points[pt].Y;
            result := True;
            exit;
        end;
    end;
    result := False;
end;

// Add the point to the points array.
procedure TRegion.AddPoint(X, Y : Integer);
var
    new_points : PPointArray;
    pt         : Integer;
begin
    // Create a new array to hold the values.
    GetMem(new_points, (num_points + 1) * SizeOf(TPoint));

    // Copy the values into the new array.
    for pt := 1 to num_points do
        new_points^[pt] := points^[pt];

    // Add the new point.
    new_points^[num_points + 1].X := X;
    new_points^[num_points + 1].Y := Y;

    // Free the previous list.
    if (num_points > 0) then FreeMem(points);

    // Make points point to the new array;
    points := new_points;

    // Update num_points.
    num_points := num_points + 1;
end;

// Copy the first point into the last position to close
// the polygonal region.
procedure TRegion.FinishRegion;
begin
    AddPoint(points[1].X, points[1].Y);
end;

// Draw the region.
procedure TRegion.DrawRegion(cvs : TCanvas; colors : array of TColor; show_region_numbers : Boolean);
begin
    // Do nothing if we have no points.
    if (num_points < 2) then exit;

    // Fill the region with its color.
    cvs.Brush.Color := colors[color_number];
    cvs.Polygon(Slice(points^, num_points));

    // Display the region number.
    if (show_region_numbers) then ShowNodeNumber(cvs);
end;

// Read this region's data from the file.
procedure TRegion.LoadRegion(var in_file : TextFile);
var
    num_pts, pt, X, Y : Integer;
begin
    // Read the number of points in this region.
    Readln(in_file, num_pts);
    num_points := 0;

    // Read the points.
    for pt := 1 to num_pts do
    begin
        Readln(in_file, X, Y);
        AddPoint(X, Y);
    end;
end;

// Write this region's data into the file.
procedure TRegion.SaveRegion(var out_file : TextFile);
var
    pt : Integer;
begin
    // Save the number of points in this region.
    Writeln(out_file, num_points);

    // Save the points.
    for pt := 1 to num_points do
        Writeln(out_file, points[pt].X, ' ', points[pt].Y);
end;

// Return True if the region contains this line segment.
function TRegion.HasLine(a1, b1, a2, b2 : Integer) : Boolean;
var
    i, x1, y1, x2, y2 : Integer;
begin
    // Assume we will find it.
    Result := True;

    x1 := points^[1].X;
    y1 := points^[1].Y;
    for i := 2 to num_points do
    begin
        x2 := points^[i].X;
        y2 := points^[i].Y;
        if ((x1 = a1) and (y1 = b1) and
            (x2 = a2) and (y2 = b2)) then exit;
        if ((x2 = a1) and (y2 = b1) and
            (x1 = a2) and (y1 = b2)) then exit;

        x1 := x2;
        y1 := y2;
    end;

    // We didn't find the line.
    Result := False;
end;

// See if this region is adjacent to the indicated one.
// If so, add the regions to each other's adjacency lists.
procedure TRegion.CheckNeighbor(rgn : TRegion);
var
    i, x1, y1, x2, y2 : Integer;
begin
    x1 := points^[1].X;
    y1 := points^[1].Y;
    for i := 2 to num_points do
    begin
        x2 := points^[i].X;
        y2 := points^[i].Y;
        if (rgn.HasLine(x1, y1, x2, y2)) then
        begin
            // They are neighbors. Add them to each other's
            // adjacency lists.
            neighbors.Add(rgn);
            rgn.neighbors.Add(Self);
            exit;
        end;
        x1 := x2;
        y1 := y2;
    end;
end;

// Remove this neighbor from our neighbors list.
procedure TRegion.RemoveNeighbor(target : TRegion);
var
    nbr_num : Integer;
    nbr     : TRegion;
begin
    // Find the neighbor.
    for nbr_num := 0 to neighbors.Count - 1 do
    begin
        nbr := neighbors.Items[nbr_num];
        if (nbr = target) then
        begin
            // We found it. Remove it.
            neighbors.Delete(nbr_num);
            exit;
        end;
    end;
end;

// Remove the node from the graph.
procedure TRegion.RemoveFromGraph;
var
    nbr_num : Integer;
    nbr     : TRegion;
begin
    // Remove links from neighbors to here.
    for nbr_num := 0 to neighbors.Count - 1 do
    begin
        nbr := neighbors.Items[nbr_num];
        nbr.RemoveNeighbor(Self);
    end;
end;

// Find two mutually non-adjacent neighbors with
// degree <= 7, if they exist.
function TRegion.FindNonAdjacents(var n1, n2 : TRegion) : Boolean;
var
    n1_num, n2_num, n3_num : Integer;
    n3                     : TRegion;
    non_adjacent           : Boolean;
begin
    // If this node has more than 5 neighbors, it won't do.
    if (neighbors.Count > 5) then
    begin
        result := False;
        exit;
    end;

    for n1_num := 0 to neighbors.Count - 2 do
    begin
        // See if n1 has degree <= 7.
        n1 := neighbors.Items[n1_num];
        if (n1.neighbors.Count <= 7) then
        begin
            // n1 has degree <= 7. Find another.
            for n2_num := n1_num + 1 to neighbors.Count - 1 do
            begin
                // See if n2 has degree <= 7.
                n2 := neighbors.Items[n2_num];
                if (n2.neighbors.Count <= 7) then
                begin
                    // See if n1 and n2 are non-adjacent.
                    non_adjacent := True;
                    for n3_num := 0 to n1.neighbors.Count - 1 do
                    begin
                        n3 := n1.neighbors.Items[n3_num];
                        if (n3 = n2) then
                        begin
                            // They are adjacent.
                            non_adjacent := False;
                            break;
                        end;
                    end;

                    // If the nodes are non-adjacent,
                    // we are done.
                    if (non_adjacent) then
                    begin
                        result := True;
                        exit;
                    end;
                end; // End if (n2.neighbors.Count <= 7) then
            end; // End for n2_num = n1_num + 1 to neighbors.Count - 1 do
        end; // End if (n1.neighbors.Count <= 7) then
    end; // End for n1_num = 0 to neighbors.Count - 2 do

    // We did not find a usable pair of nodes.
    result := False;
end;

// If this region is not yet in our neighbor list, add it
// and add us to its list.
procedure TRegion.AddNeighbor(nbr : TRegion);
var
    n1_num : Integer;
    n1     : TRegion;
begin
    // Examine our neighbors.
    for n1_num := 0 to neighbors.Count - 1 do
    begin
        n1 := neighbors.Items[n1_num];
        // If the node is in the neighbors list, do nothing.
        if (n1 = nbr) then exit;
    end;

    // Update the neighbor lists.
    neighbors.Add(nbr);
    nbr.neighbors.Add(Self);
end;

// Associate this node with the target. Copy this node's
// neighbors into target's neighbor list.
procedure TRegion.AssociateWith(target : TRegion);
var
    n1_num : Integer;
    n1     : TRegion;
begin
    // Examine all neighbors.
    for n1_num := 0 to neighbors.Count - 1 do
    begin
        n1 := neighbors.Items[n1_num];
        if (n1 <> target) then
        begin
            // Add n1 to target's neighbor list and
            // vice versa.
            n1.AddNeighbor(target);
        end;
    end;
end;

// Display a message listing this region's neighbors.
function TRegion.ListNeighbors : String;
var
    i   : Integer;
    nbr : TRegion;
begin
    result := Format('Region %d: ', [number]);
    for i := 0 to neighbors.Count - 1 do
    begin
        nbr := neighbors.Items[i];
        result := result + Format('%d ', [nbr.number]);
    end;
    result := result + #13#10;
end;

// Display the region's number.
procedure TRegion.ShowNodeNumber(cvs : TCanvas);
var
    i, X, Y : Integer;
    txt     : String;
begin
    X := 0;
    Y := 0;
    // Find average region coordinate.
    for i := 1 to num_points do
    begin
        X := X + points[i].X;
        Y := Y + points[i].Y;
    end;

    txt := Format('%d', [number]);
    X := Round(X / num_points - cvs.TextWidth(txt) / 2);
    Y := Round(Y / num_points - cvs.TextHeight(txt) / 2);
    cvs.TextOut(X, Y, txt);
end;

// ******************
// * TMapForm class *
// ******************
// Exit.
procedure TMapForm.mnuFileExitClick(Sender: TObject);
begin
    Close;
end;

procedure TMapForm.FormCreate(Sender: TObject);
begin
    // Draw in black.
    Canvas.Pen.Color := clRed;

    // Initialize the colors.
    colors[0] := clWhite;   // Color for unassigned regions.
    colors[1] := clLime;
    colors[2] := clBlue;
    colors[3] := clYellow;
    colors[4] := clFuchsia;
    colors[5] := clMaroon;

    // Start with an empty regions list.
    regions := TList.Create;
end;

// If there is an existing point within SNAP pixels in
// both the X and Y directions, change X and Y to its
// coordinates.
procedure TMapForm.SnapTo(var X, Y : Integer);
var
    i   : Integer;
    rgn : TRegion;
begin
    for i := 0 to regions.Count - 1 do
    begin
         rgn := regions.Items[i];
        if (rgn.SnapTo(X, Y)) then exit;
    end;
end;

// Create a new region, add it to the regions list,
// and set its number value. Return the new region.
function TMapForm.AddRegion : TRegion;
begin
    // Create the new region.
    result := TRegion.Create;
    result.number := regions.Count;
    regions.Add(result);
end;

// Start selecting a region's point.
procedure TMapForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
    SnapTo(X, Y);

    // If we are currently drawing a region and the user
    // clicks the right button, finish the region.
    if ((Button = mbRight) and
        (drawing_region <> nil)) then
    begin
        // Finish the region we're drawing.
        // Erase the last rubberband line.
        Canvas.Pen.Mode := pmXor;
        Canvas.MoveTo(last_x, last_y);
        Canvas.LineTo(cur_x, cur_y);
        Canvas.Pen.Mode := pmCopy;

        // Finish the current drawing region.
        drawing_region.FinishRegion;

        // Draw the closing segment.
        Canvas.MoveTo(last_x, last_y);
        cur_x := drawing_region.points^[
            drawing_region.num_points].X;
        cur_y := drawing_region.points^[
            drawing_region.num_points].Y;
        Canvas.LineTo(cur_x, cur_y);

        drawing_region := nil;
        exit;
    end;

    // If we have not started a region, start one.
    if (drawing_region = nil) then
    begin
        // Create a new region.
        drawing_region := AddRegion;
    end else begin
        // Redraw the last line in black.
        Canvas.MoveTo(last_x, last_y);
        Canvas.LineTo(cur_x, cur_y);
    end;

    // Add the new point to the current region.
    drawing_region.AddPoint(X, Y);

    // Record the next point's starting position.
    last_x := X;
    last_y := Y;
    cur_x := X;
    cur_y := Y;
end;

// If we are currently drawing, draw a rubberband line to
// show where the next point will be.
procedure TMapForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
    // Do nothing if we are not drawing.
    if (drawing_region = nil) then exit;

    // Erase the previous rubberband line.
    Canvas.Pen.Mode := pmXor;
    Canvas.MoveTo(last_x, last_y);
    Canvas.LineTo(cur_x, cur_y);

    // Record the new point's position.
    SnapTo(X, Y);
    cur_x := X;
    cur_y := Y;

    // Draw the new rubberband line.
    Canvas.MoveTo(last_x, last_y);
    Canvas.LineTo(cur_x, cur_y);
    Canvas.Pen.Mode := pmCopy;
end;

// Redraw the map.
procedure TMapForm.FormPaint(Sender: TObject);
var
    i                   : Integer;
    rgn                 : TRegion;
    show_region_numbers : Boolean;
begin
    show_region_numbers := mnuShowRegionNumbers.Checked;
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        rgn.DrawRegion(Canvas, colors, show_region_numbers);
    end;
end;

// Load a map file.
procedure TMapForm.LoadMap(file_name : String);
var
    in_file          : TextFile;
    num_regions, rgn : Integer;
    new_rgn          : TRegion;
begin
    AssignFile(in_file, file_name);
    Reset(in_file);

    // See how many regions there are.
    Readln(in_file, num_regions);

    // Read the regions.
    regions.Clear;
    for rgn := 1 to num_regions do
    begin
        new_rgn := AddRegion;

        // Make the new region read itself from the file.
        new_rgn.LoadRegion(in_file);
    end;

    // Close the map file.
    CloseFile(in_file);
    drawing_region := nil;
end;

// Load a map file and draw it.
procedure TMapForm.mnuFileOpenClick(Sender: TObject);
begin
    if (dlgOpen.Execute) then
    begin
        LoadMap(dlgOpen.FileName);
        Caption := 'Color5 [' +
            ExtractFileName(dlgOpen.FileName) + ']';
        Refresh;
    end;
end;

// Create the adjacency lists for the regions.
procedure TMapForm.FindNeighbors;
var
    i1, i2     : Integer;
    rgn1, rgn2 : TRegion;
begin
    // Compare each region to every other and see if they
    // are adjacent.
    for i1 := 0 to regions.Count - 2 do
    begin
        rgn1 := regions.Items[i1];
        for i2 := i1 + 1 to regions.Count - 1 do
        begin
            rgn2 := regions.Items[i2];
            rgn1.CheckNeighbor(rgn2);
        end;
    end;

    // Display the neighbor lists if desired.
    if (mnuShowNeighborLists.Checked) then
        ShowNeighborLists;
end;

// Five-color the map.
procedure TMapForm.AssignColors;
var
    orig_regions, stack, nbrs : TList;
    i, rgn_num, n1_num        : Integer;
    rgn, n1, n2               : TRegion;
    color_used                : array [1..5] of Boolean;
    obj                       : TObject;
begin
    // Prepare the lists we need.
    orig_regions := TList.Create;
    stack := TList.Create;

    // Save a copy of the region list because we will
    // mess it up. Also reset the region's colors and
    // clear their adjacency lists.
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        orig_regions.Add(rgn);

        rgn.color_number := 0;
        rgn.neighbors.Clear;
    end;

    // Make the adjacency lists.
    FindNeighbors;

    // Push regions onto the stack.
    while (regions.Count > 0) do
    begin
        // Look for a region with degree < 5.
        rgn := nil;
        for rgn_num := 0 to regions.Count - 1 do
        begin
            rgn := regions.Items[rgn_num];
            if (rgn.neighbors.Count < 5) then break;
        end;

        // If we found node with fewer than 5 neighbors,
        // add it and its neighbor list to the stack.
        if (rgn_num < regions.Count) then
        begin
            // Push rgn and its neighor list onto the stack.
            stack.Add(rgn);
            stack.Add(rgn.Neighbors);

            // Remove rgn from the graph.
            rgn.RemoveFromGraph;
            regions.Delete(rgn_num);
        end else begin
            // There is no node with degree < 5. Search
            // for one with degree = 5 and 2 non-adjacent
            // nodes n1 and n2 with degree <= 7.
            for rgn_num := 0 to regions.Count - 1 do
            begin
                rgn := regions.Items[rgn_num];
                if (rgn.FindNonAdjacents(n1, n2)) then
                    break;
            end;

            // If we still did not find one, something
            // is wrong.
            if (rgn_num >= regions.Count) then
            begin
                ShowMessage('Error finding node to remove.');
                break;
            end;

            // Save rgn and its adjacency list.
            stack.Add(rgn);
            stack.Add(rgn.neighbors);

            // Remove rgn from the graph.
            rgn.RemoveFromGraph;
            regions.Delete(rgn_num);

            // Associate n1 and n2.
            n1.AssociateWith(n2);

            // Push n1 and n2 onto the stack.
            stack.Add(n1);
            stack.Add(n2);

            // Remove n1 from the graph.
            n1.RemoveFromGraph;

            // Find n1 in the region list.
            for n1_num := 0 to regions.Count - 1 do
            begin
                n2 := regions.Items[n1_num];
                if (n1 = n2) then
                begin
                    regions.Delete(n1_num);
                    break;
                end;
            end;
        end; // End if we did not find a node with degree < 5.
    end; // End while (regions.Count > 0) do

    // The graph is empty. Produce the coloring.

    // Pop regions off the stack and color them.
    while (stack.Count > 0) do
    begin
        // See if the next item in the stack is a
        // neighbor list or a region.
        obj := stack.Items[stack.Count - 1];
        if (obj.ClassNameIs('TList')) then
        begin
            // Get the neighbor list.
            nbrs := stack.Items[stack.Count - 1];
            stack.Delete(stack.Count - 1);

            // Get the corresponding region.
            rgn := stack.Items[stack.Count - 1];
            stack.Delete(stack.Count - 1);

            // Assign rgn a color different from those
            // used by its neighbors.
            for i := 1 To 5 do color_used[i] := False;
            for n1_num := 0 to nbrs.Count - 1 do
            begin
                n1 := nbrs.Items[n1_num];
                color_used[n1.color_number] := True;
            end;

            // See which color is left.
            for i := 1 to 5 do
                if (not color_used[i]) then break;

            // If we did not find an unused color,
            // something is wrong.
            if (i > 5) then
                ShowMessage('Error finding color for node.')
            else
                rgn.color_number := i;

            // End if the next item in the stack
            // is a neighbor list.
        end else begin
            // The next item in the stack is a region.
            // Get the region.
            n1 := stack.Items[stack.Count - 1];
            stack.Delete(stack.Count - 1);

            // Get the associated region.
            n2 := stack.Items[stack.Count - 1];
            stack.Delete(stack.Count - 1);

            // Assign n2 the same color as n1.
            n2.color_number := n1.color_number;
        end; // End if neighbor list/node ... else node/node ...
    end; // End while (stack.Count > 0) do

    // Restore the original region list.
    regions.Destroy;
    regions := orig_regions;
end;

// Call AssignColors to five-color the map.
procedure TMapForm.mnuColorNodesClick(Sender: TObject);
begin
    Cursor := crHourglass;

    // Assign the colors and redraw the map.
    AssignColors;
    Refresh;

    Cursor := crDefault;
end;

// Save the map.
procedure TMapForm.mnuFileSaveAsClick(Sender: TObject);
begin
    if (dlgSaveAs.Execute) then
        SaveMap(dlgSaveAs.FileName);
end;

// Load a map file.
procedure TMapForm.SaveMap(file_name : String);
var
    out_file : TextFile;
    i        : Integer;
    rgn      : TRegion;
begin
    AssignFile(out_file, file_name);
    Rewrite(out_file);

    // Save the number of regions.
    Writeln(out_file, regions.Count);

    // Save the regions.
    for i := 0 to regions.Count - 1 do
    begin
        // Make the new region write itself into the file.
        rgn := regions.Items[i];
        rgn.SaveRegion(out_file);
    end;

    // Close the map file.
    CloseFile(out_file);
end;

// Start a fresh map.
procedure TMapForm.mnuFileNewClick(Sender: TObject);
begin
    regions.Clear;
    drawing_region := nil;
    Refresh;
end;

procedure TMapForm.ShowNeighborLists;
var
    rgn : TRegion;
    i   : Integer;
    txt : String;
begin
    txt := '';
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        txt := txt + rgn.ListNeighbors;
    end;
    ShowMessage(txt);
end;

// Show the node numbers.
procedure TMapForm.ShowNodeNumbers;
var
    rgn : TRegion;
    i   : Integer;
begin
    for i := 0 to regions.Count - 1 do
    begin
        rgn := regions.Items[i];
        rgn.ShowNodeNumber(Canvas);
    end;
end;

// Toggle the Show Region Numbers option.
procedure TMapForm.mnuShowRegionNumbersClick(Sender: TObject);
begin
    mnuShowRegionNumbers.Checked :=
        (not mnuShowRegionNumbers.Checked);
    Refresh;
end;

// Toggle the Show Neighbor Lists option.
procedure TMapForm.mnuShowNeighborListsClick(Sender: TObject);
begin
    mnuShowNeighborLists.Checked :=
        (not mnuShowNeighborLists.Checked);
end;

end.
