{
  Name    :   DelphiBankCommonDM.pas

  Purpose :   Anscestor datamodule for all other server datamodules.
              Provides field name constants, database setup code,
               a dataset to variant array conversion function and
               client key verification.

  Usage   :   This source code is supplied as example code with Delphi Informant
               magazine.
              The author makes no guarantee for its correctness or fitness
               for any purpose and cannot be held responsible for any loss
               or damage that may occur through its use or misuse.
              You may copy, modify or use any part of this source code
               provided the author is credited for his work.

  Author  :   Paul M. Fairhurst, September 1998. paul@c-s-c.demon.co.uk
}
unit DelphiBankCommonDM;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DB, DBTables;

const
  cTableAccounts            : String = 'ACCOUNTS';
  cTableAccountTypes        : String = 'ACCTTYPE';
  cTableAccountCustomers    : String = 'CUSTOMERS';
  cTableAccountTransactions : String = 'TRANSACTIONS';

  cFieldATID            : String = 'AT_ID';
  cFieldATName          : String = 'AT_NAME';
  cFieldATInterestRate  : String = 'AT_INTEREST_RATE';
  cFieldATDefMinBalance : String = 'AT_DEF_MINIMUM_BAL';

  cFieldAcctID             : String = 'ACCT_ID';
  cFieldAcctMinBalance     : String = 'ACCT_MIN_BALANCE';
  cFieldAcctBalance        : String = 'ACCT_BALANCE';
  cFieldAcctLastStatement  : String = 'ACCT_LAST_STATEMENT';

  cFieldCustID       : String = 'CUST_ID';
  cFieldCustName     : String = 'CUST_NAME';
  cFieldCustLogonID  : String = 'CUST_LOGON_ID';
  cFieldCustPassword : String = 'CUST_PASSWORD';
  cFieldCustNotes    : String = 'CUST_NOTES';

  cFieldTrnID        : String = 'TRN_ID';
  cFieldTrnRef       : String = 'TRN_REF';
  cFieldTrnAmount    : String = 'TRN_AMOUNT';
  cFieldTrnTimestamp : String = 'TRN_TIMESTAMP';

type
  TdmDelphiBankCommon = class(TDataModule)
  private
    { Private declarations }
  public
    { Public declarations }

    { Creates and populates a 2D variant array from a dataset. Indexing starts
       at [0, 0]. Row 0 (i.e. [0..m, 0] contains the column display label,
       everything else is field data from the result set.
    }
    procedure CreateVarArrayFromDataset(var varResultSet: OleVariant; ADataset : TDataset);

    { Create a new client key. }
    function  CreateClientKey : Integer;

    { Checks if a client key is valid. }
    function  IsClientKeyValid(ClientKey : Integer; var strDebug : WideString) : Boolean;

    procedure SetUpDB(ADatabase : TDatabase);
  end;

implementation

{$R *.DFM}

uses Math, Mtx;

const
  cDelphiBankClientKeys : WideString = 'DelphiBankServer3ClientKeys';
  cClientKeyTimeout : Integer = 300; { 5 Minutes (in Seconds) }

{ ---------------------------------------------------------------------------- }
procedure TdmDelphiBankCommon.CreateVarArrayFromDataset(var varResultSet: OleVariant;
                                                          ADataset : TDataset);
var
  m : Integer;
  nRecords, nColumns, nCurRec : Integer;
begin
  nRecords := -1;
  nColumns := -1;

  try
    { Create the array... }
    { Set size to 0..m-1 where m equals the number of columns. }
    nColumns := Max(0, ADataset.FieldCount-1);

    { Each item is an array of size (0..n) where n equals the number of records.
      Entry 0 is where we store the column name.
    }
    nRecords := Max(0, ADataset.RecordCount);

    varResultSet := VarArrayCreate([0, nColumns, 0, nRecords], varVariant);

    for m := 0 to nColumns do
      varResultSet[m, 0] := ADataset.Fields[m].DisplayLabel;

    { Populate from result set. }
    ADataset.First;
    nCurRec := 1; { Current record number. }
    while not ADataset.Eof do begin
      { Put in field values. }
      for m := 0 to nColumns do
        varResultSet[m, nCurRec] := ADataset.Fields[m].Value;

      ADataset.Next;
      Inc(nCurRec);
    end;
  except
    on E: Exception do
      raise Exception.Create('CreateVarArrayFromDataset() - ' + IntToStr(nRecords) +
                             ' rec,' + IntToStr(nColumns) + 'cols,' + E.Message);
  end;
end;

{ ---------------------------------------------------------------------------- }
function TdmDelphiBankCommon.CreateClientKey : Integer;
var
  spgClientKeys  : ISharedPropertyGroup;
  spClientKey    : ISharedProperty;
  wsPropertyName : WideString;
  bAlreadyExists : WordBool;
begin
  Result := 0; { Not needed, but stops compiler complaints. }

  try
    { Create property group. }
    spgClientKeys := CreateSharedPropertyGroup(cDelphiBankClientKeys);
    if not Assigned(spgClientKeys) then
      raise Exception.Create('Could not create client key property group');

    { Create the property. }
    bAlreadyExists := False;
    while not bAlreadyExists do begin
      Result := GetTickCount;
      wsPropertyName := 'CK' + IntToStr(Result);
      spClientKey := spgClientKeys.CreateProperty(wsPropertyName, bAlreadyExists);
    end;

    if not Assigned(spClientKey) then
      raise Exception.Create('Could not create client key');

    { Set the value. }
    spClientKey.Value := Result;
  except
    on E: Exception do
      raise Exception.Create('TdmDelphiBankCommon.CreateClientKey() - ' + E.Message);
  end;
end;

{ ---------------------------------------------------------------------------- }
function TdmDelphiBankCommon.IsClientKeyValid(ClientKey : Integer; var strDebug : WideString) : Boolean;
var
  spgClientKeys  : ISharedPropertyGroup;
  spClientKey    : ISharedProperty;
  nTickCountNow  : Integer;
begin
  Result   := False;
  strDebug := 'Connection has timed-out. Please logon again.';

  try
    { Create property group. }
    spgClientKeys := CreateSharedPropertyGroup(cDelphiBankClientKeys);
    if not Assigned(spgClientKeys) then
      raise Exception.Create('Could not create client key property group');

    { Look for the property. }
    spClientKey := spgClientKeys.PropertyByName['CK' + IntToStr(ClientKey)];
    if not Assigned(spClientKey) then
      Exit;

    { Check value. }
    nTickCountNow := GetTickCount;
    if (Abs(nTickCountNow) - Abs(spClientKey.Value)) <= (cClientKeyTimeout * 1000) then begin
      Result   := True;
      strDebug := '';
    end;
  except
    on E: Exception do begin
      strDebug := 'TdmDelphiBankCommon.IsClientKeyValid() - Client Key of "' +
                  'CK' + IntToStr(ClientKey) + '" -> ' + E.Message;
      raise Exception.Create(strDebug);
    end;
  end;
end;

{ ---------------------------------------------------------------------------- }
procedure TdmDelphiBankCommon.SetUpDB(ADatabase : TDatabase);
begin
  with ADatabase do begin
    Connected   := False;
    AliasName   := 'PDX The Delphi Bank'; { Must be in BDE. }
    LoginPrompt := False;
    Params.Clear;
    Params.Add('USER NAME=delphibank_dbo'); { Not needed for Paradox... }
    Params.Add('PASSWORD=delphi');
  end;
end;

end.
