{Copyright:      Hagen Reddmann  mailto:HaReddmann@AOL.COM
 Author:         Hagen Reddmann
 Remarks:        freeware, but this Copyright must be included
 known Problems: none
 Version:        2.1,  Part I from Delphi Encryption Compendium
                 Delphi 2-4, designed and testet under D3
 Description:    included the Designtime Components
                 THashManager for the THash-Administration
                 TCipherManager for the TCipher-Administration

 Version 2.1
                 TCipherManager: Methods added
                   EncodeBuffer(), DecodeBuffer()
                   EncodeString(), DecodeString()

}
unit HCMngr;

interface

uses Classes, Hash, Cipher;

type
  THashManager = class(TComponent)
  private
    FHashClass: THashClass;
    FHash: THash;
    FDigest: Pointer;
    function GetInfo(Index: Integer): String;
    procedure SetInfo(Index: Integer; Value: String);
    function GetHashClass: THashClass;
    procedure SetHashClass(Value: THashClass);
    function GetHash: THash;
    function GetAlgorithm: String;
    procedure SetAlgorithm(Value: String);
    function GetDigestStr(Index: Integer): String;
    function GetDigestSize: Integer;
    procedure ReadHash(Reader: TReader);
    procedure WriteHash(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    destructor Destroy; override;
    procedure CalcBuffer(const Buffer; BufferSize: Integer);
    procedure CalcStream(const Stream: TStream; StreamSize: Integer);
    procedure CalcString(const Data: String);
    procedure CalcFile(const FileName: String);
    class function RegisterHash(const AHash: THashClass; const AName, ADescription: String): Boolean;
    class function UnregisterHash(const AHash: THashClass): Boolean;
    class function GetHashList(const Strings: TStrings): Boolean;
    property Digest: Pointer read FDigest;
    property DigestSize: Integer read GetDigestSize;
    property DigestKeyStr: String index -1 read GetDigestStr;
    property DigestString: String index  0 read GetDigestStr;
    property DigestBase16: String index 16 read GetDigestStr;
    property DigestBase64: String index 64 read GetDigestStr;
    property HashClass: THashClass read GetHashClass write SetHashClass;
    property Hash: THash read GetHash;
  published
    property Algorithm: String read GetAlgorithm write SetAlgorithm stored False;
    property Description: String index 0 read GetInfo write SetInfo stored False;
  end;

  TCipherManager = class(TComponent)
  private
    FHashManager: THashManager;
    FCipherClass: TCipherClass;
    FCipher: TCipher;
    FMode: TCipherMode;
    FIncludeHashKey: Boolean;
    procedure SetHashManager(Value: THashManager);
    procedure SetMode(Value: TCipherMode);
    procedure SetIncludeHashKey(Value: Boolean);
    function GetCipher: TCipher;
    function GetCipherClass: TCipherClass;
    procedure SetCipherClass(Value: TCipherClass);
    function GetAlgorithm: String;
    procedure SetAlgorithm(Value: String);
    function GetInfo(Index: Integer): String;
    procedure SetInfo(Index: Integer; Value: String);
    procedure ReadCipher(Reader: TReader);
    procedure WriteCipher(Writer: TWriter);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DefineProperties(Filer: TFiler); override;
  public
    destructor Destroy; override;
    procedure InitKey(const Key: String; IVector: Pointer);
    procedure EncodeFile(const Source, Dest: String);
    procedure DecodeFile(const Source, Dest: String);
    procedure EncodeStream(Source, Dest: TStream; DataSize: Integer);
    procedure DecodeStream(Source, Dest: TStream; DataSize: Integer);
    procedure EncodeBuffer(const Source; var Dest; DataSize: Integer);
    procedure DecodeBuffer(const Source; var Dest; DataSize: Integer);
    function EncodeString(const Source: String): String;
    function DecodeString(const Source: String): String;

    class function RegisterCipher(const ACipher: TCipherClass; const AName, ADescription: String): Boolean;
    class function UnregisterCipher(const ACipher: TCipherClass): Boolean;
    class function GetCipherList(const Strings: TStrings): Boolean;
    property Cipher: TCipher read GetCipher;
    property CipherClass: TCipherClass read GetCipherClass write SetCipherClass;
  published
    property Algorithm: String read GetAlgorithm write SetAlgorithm stored False;
    property Description: String index 0 read GetInfo write SetInfo stored False;
    property Mode: TCipherMode read FMode write SetMode default cmCBC;
    property HashManager: THashManager read FHashManager write SetHashManager;
    property IncludeHashKey: Boolean read FIncludeHashKey write SetIncludeHashKey;
  end;

procedure Register;

implementation

uses SysUtils, Windows, DsgnIntf;

const
  FHashList: TStringList = nil;
  FCipherList: TStringList = nil;

procedure THashManager.ReadHash(Reader: TReader);
var
  CN: String;
  I: Integer;
begin
  CN := Reader.ReadString;
  for I := 0 to FHashList.Count-1 do
    if CN = THashClass(FHashList.Objects[I]).ClassName then
    begin
      SetHashClass(THashClass(FHashList.Objects[I]));
      Exit;
    end;
end;

procedure THashManager.WriteHash(Writer: TWriter);
begin
  Writer.WriteString(HashClass.ClassName);
end;

procedure THashManager.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Hash', ReadHash, WriteHash,
                        (FHashClass <> nil) and (FHashClass <> THash_CRC32));
end;

function THashManager.GetInfo(Index: Integer): String;
begin
  Result := '';
  case Index of
    0: begin
         Index := FHashList.IndexOfObject(Pointer(HashClass));
         if Index >= 0 then
         begin
           Result := Trim(FHashList[Index]);
           Delete(Result, 1, Pos('=', Result));
         end;
         if Length(Result) > 0 then Result := Result + ', ';
         Result := Result + IntToStr(DigestSize * 8) + 'bit Digestsize';
       end;
  end;
end;

procedure THashManager.SetInfo(Index: Integer; Value: String);
begin
end;

function THashManager.GetDigestStr(Index: Integer): String;
begin
  Result := '';
  if FDigest = nil then Exit;
  if Index = 0 then Index := DefaultDigestStringFormat;
  case Index of
    16: Result := StrToBase16(PChar(FDigest), DigestSize);
    64: Result := StrToBase64(PChar(FDigest), DigestSize);
  else
    begin
      SetLength(Result, DigestSize);
      Move(FDigest^, PChar(Result)^, DigestSize);
    end;
  end;
end;

function THashManager.GetDigestSize: Integer;
begin
  Result := HashClass.DigestKeySize;
end;

function THashManager.GetHashClass: THashClass;
begin
  if FHashClass = nil then SetHashClass(THash_CRC32);
  Result := FHashClass;
end;

function THashManager.GetHash: THash;
begin
  if FHash = nil then FHash := HashClass.Create;
  Result := FHash;
end;

procedure THashManager.SetHashClass(Value: THashClass);
begin
  if Value <> FHashClass then
  begin
    if Value = nil then Value := THash_CRC32;
    FHashClass := Value;
    FHash.Free;
    FHash := nil;
    ReallocMem(FDigest, DigestSize);
    FillChar(FDigest^, DigestSize, 0);
  end;
end;

function THashManager.GetAlgorithm: String;
var
  I: Integer;
begin
  Result := '';
  I := FHashList.IndexOfObject(Pointer(HashClass));
  if I >= 0 then Result := FHashList.Names[I];
end;

procedure THashManager.SetAlgorithm(Value: String);
var
  I: Integer;
begin
  I := FHashList.IndexOfName(Value);
  if I >= 0 then SetHashClass(Pointer(FHashList.Objects[I]))
    else SetHashClass(nil);
end;

destructor THashManager.Destroy;
begin
  FHash.Free;
  FHash := nil;
  inherited Destroy;
end;

procedure THashManager.CalcBuffer(const Buffer; BufferSize: Integer);
begin
  HashClass.CalcBuffer(FDigest, Buffer, BufferSize);
end;

procedure THashManager.CalcStream(const Stream: TStream; StreamSize: Integer);
begin
  HashClass.CalcStream(FDigest, Stream, StreamSize);
end;

procedure THashManager.CalcString(const Data: String);
begin
  HashClass.CalcString(FDigest, Data);
end;

procedure THashManager.CalcFile(const FileName: String);
begin
  HashClass.CalcFile(FDigest, FileName);
end;

class function THashManager.RegisterHash(const AHash: THashClass; const AName, ADescription: String): Boolean;
var
  I: Integer;
  S: String;
begin
  Result := False;
  if AHash = nil then Exit;
  S := Trim(AName);
  if S = '' then
  begin
    S := AHash.ClassName;
    if S[1] = 'T' then Delete(S, 1, 1);
    I := Pos('_', S);
    if I > 0 then Delete(S, 1, I);
  end;
  S := S + '=' + ADescription;
  I := FHashList.IndexOfObject(Pointer(AHash));
  if I < 0 then FHashList.AddObject(S, Pointer(AHash))
    else FHashList[I] := S;
  Result := True;
end;

class function THashManager.UnregisterHash(const AHash: THashClass): Boolean;
var
  I: Integer;
begin
  Result := False;
  repeat
    I := FHashList.IndexOfObject(Pointer(AHash));
    if I < 0 then Break;
    Result := True;
    FHashList.Delete(I);
  until False;
end;

class function THashManager.GetHashList(const Strings: TStrings): Boolean;
var
  I: Integer;
begin
  Result := False;
  if Strings = nil then Exit;
  Strings.Clear;
  for I := 0 to FHashList.Count -1 do
    Strings.AddObject(FHashList.Names[I], FHashList.Objects[I]);
end;

procedure TCipherManager.SetHashManager(Value: THashManager);
begin
  if Value <> FHashManager then
  begin
    FHashManager := Value;
    if FHashManager <> nil then FHashManager.FreeNotification(Self);
  end;
end;

procedure TCipherManager.SetMode(Value: TCipherMode);
begin
  FMode := Value;
  if FCipher <> nil then FCipher.Mode := FMode;
end;

procedure TCipherManager.SetIncludeHashKey(Value: Boolean);
begin
  FIncludeHashKey := Value;
  if FCipher <> nil then FCipher.IncludeHashKey := FIncludeHashKey;
end;

function TCipherManager.GetCipher: TCipher;
begin
  if FCipher = nil then
  begin
    FCipher := CipherClass.Create;
    FCipher.Mode := FMode;
    FCipher.IncludeHashKey := FIncludeHashKey;
  end;  
  Result := FCipher;
end;

function TCipherManager.GetCipherClass: TCipherClass;
begin
  if FCipherClass = nil then FCipherClass := TCipher_Blowfish;
  Result := FCipherClass;
end;

procedure TCipherManager.SetCipherClass(Value: TCipherClass);
begin
  if Value <> CipherClass then
  begin
    FCipher.Free;
    FCipher := nil;
    FCipherClass := Value;
  end;
end;

function TCipherManager.GetAlgorithm: String;
var
  I: Integer;
begin
  Result := '';
  I := FCipherList.IndexOfObject(Pointer(CipherClass));
  if I >= 0 then Result := FCipherList.Names[I];
end;

procedure TCipherManager.SetAlgorithm(Value: String);
var
  I: Integer;
begin
  I := FCipherList.IndexOfName(Value);
  if I >= 0 then SetCipherClass(Pointer(FCipherList.Objects[I]))
    else SetCipherClass(nil);
end;

function TCipherManager.GetInfo(Index: Integer): String;
begin
  Result := '';
  case Index of
    0: begin
         Index := FCipherList.IndexOfObject(Pointer(CipherClass));
         if Index >= 0 then
         begin
           Result := Trim(FCipherList[Index]);
           Delete(Result, 1, Pos('=', Result));
         end;
         if Length(Result) > 0 then Result := Result + ', ';
         Result := Result + IntToStr(CipherClass.MaxKeySize * 8) + 'bit Key';
       end;
  end;
end;

procedure TCipherManager.SetInfo(Index: Integer; Value: String);
begin
end;

procedure TCipherManager.ReadCipher(Reader: TReader);
var
  CN: String;
  I: Integer;
begin
  CN := Reader.ReadString;
  for I := 0 to FCipherList.Count-1 do
    if CN = TCipherClass(FCipherList.Objects[I]).ClassName then
    begin
      SetCipherClass(TCipherClass(FCipherList.Objects[I]));
      Exit;
    end;
end;

procedure TCipherManager.WriteCipher(Writer: TWriter);
begin
  Writer.WriteString(CipherClass.ClassName);
end;

procedure TCipherManager.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FHashManager) then
    HashManager := nil;
end;

procedure TCipherManager.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Cipher', ReadCipher, WriteCipher,
                        (FCipherClass <> nil) and (FCipherClass <> TCipher_Blowfish));
end;

destructor TCipherManager.Destroy;
begin
  FCipher.Free;
  FCipher := nil;
  inherited Destroy;
end;

procedure TCipherManager.InitKey(const Key: String; IVector: Pointer);
begin
  if FHashManager <> nil then Cipher.HashClass := FHashManager.HashClass;
  Cipher.InitKey(Key, IVector);
end;

procedure TCipherManager.EncodeFile(const Source, Dest: String);
begin
  Cipher.EncodeFile(Source, Dest);
end;

procedure TCipherManager.DecodeFile(const Source, Dest: String);
begin
  Cipher.DecodeFile(Source, Dest);
end;

procedure TCipherManager.EncodeStream(Source, Dest: TStream; DataSize: Integer);
begin
  Cipher.EncodeStream(Source, Dest, DataSize);
end;

procedure TCipherManager.DecodeStream(Source, Dest: TStream; DataSize: Integer);
begin
  Cipher.DecodeStream(Source, Dest, DataSize);
end;

procedure TCipherManager.EncodeBuffer(const Source; var Dest; DataSize: Integer);
begin
  Cipher.EncodeBuffer(Source, Dest, DataSize);
end;

procedure TCipherManager.DecodeBuffer(const Source; var Dest; DataSize: Integer);
begin
  Cipher.DecodeBuffer(Source, Dest, DataSize);
end;

function TCipherManager.EncodeString(const Source: String): String;
begin
  Result := Cipher.EncodeString(Source);
end;

function TCipherManager.DecodeString(const Source: String): String;
begin
  Result := Cipher.DecodeString(Source);
end;

class function TCipherManager.RegisterCipher(const ACipher: TCipherClass; const AName, ADescription: String): Boolean;
var
  I: Integer;
  S: String;
begin
  Result := False;
  if ACipher = nil then Exit;
  S := Trim(AName);
  if S = '' then
  begin
    S := ACipher.ClassName;
    if S[1] = 'T' then Delete(S, 1, 1);
    I := Pos('_', S);
    if I > 0 then Delete(S, 1, I);
  end;
  S := S + '=' + ADescription;
  I := FCipherList.IndexOfObject(Pointer(ACipher));
  if I < 0 then FCipherList.AddObject(S, Pointer(ACipher))
    else FCipherList[I] := S;
  Result := True;
end;

class function TCipherManager.UnregisterCipher(const ACipher: TCipherClass): Boolean;
var
  I: Integer;
begin
  Result := False;
  repeat
    I := FCipherList.IndexOfObject(Pointer(ACipher));
    if I < 0 then Break;
    Result := True;
    FCipherList.Delete(I);
  until False;
end;

class function TCipherManager.GetCipherList(const Strings: TStrings): Boolean;
var
  I: Integer;
begin
  Result := False;
  if Strings = nil then Exit;
  Strings.Clear;
  for I := 0 to FCipherList.Count -1 do
    Strings.AddObject(FCipherList.Names[I], FCipherList.Objects[I]);
end;

type
  TAlgorithmHashProperty = class(TStringProperty)
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TAlgorithmCipherProperty = class(TAlgorithmHashProperty)
    procedure GetValues(Proc: TGetStrProc); override;
  end;

function TAlgorithmHashProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paReadOnly, paValueList, paMultiSelect];
end;

procedure TAlgorithmHashProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
begin
  with FHashList do
    for I := 0 to Count-1 do Proc(Names[I]);
end;

procedure TAlgorithmCipherProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
begin
  with FCipherList do
    for I := 0 to Count-1 do Proc(Names[I]);
end;

procedure Register;
begin
  RegisterComponents('Encryption', [THashManager, TCipherManager]);
  RegisterPropertyEditor(TypeInfo(String), THashManager, 'Algorithm', TAlgorithmHashProperty);
  RegisterPropertyEditor(TypeInfo(String), TCipherManager, 'Algorithm', TAlgorithmCipherProperty);
end;

initialization
  FHashList := TStringList.Create;
  FCipherList := TStringList.Create;
  with THashManager do
  begin
    RegisterHash(THash_MD4, 'Message Digest 4', 'Hash');
    RegisterHash(THash_MD5, 'Message Digest 5', 'Hash');
    RegisterHash(THash_SHA1, 'Secure Hash Algorithm 1', 'Hash');
    RegisterHash(THash_RipeMD128, 'Ripe Message Digest 128', 'Hash');
    RegisterHash(THash_RipeMD160, 'Ripe Message Digest 160', 'Hash');
    RegisterHash(THash_RipeMD256, 'Ripe Message Digest 256', 'Hash');
    RegisterHash(THash_RipeMD320, 'Ripe Message Digest 320', 'Hash');
    RegisterHash(THash_Haval256, 'Haval-256', 'Hash');
    RegisterHash(THash_Haval224, 'Haval-224', 'Hash');
    RegisterHash(THash_Haval192, 'Haval-192', 'Hash');
    RegisterHash(THash_Haval160, 'Haval-160', 'Hash');
    RegisterHash(THash_Haval128, 'Haval-128', 'Hash');
    RegisterHash(THash_Snefru, 'Snefru-256', 'Hash');
    RegisterHash(THash_Tiger, 'Tiger', 'Hash');
    RegisterHash(THash_XOR16, 'XOR-16', 'Checksum');
    RegisterHash(THash_XOR32, 'XOR-32', 'Checksum');
    RegisterHash(THash_CRC32, 'CRC-32', 'Checksum');
  end;
  with TCipherManager do
  begin
    RegisterCipher(TCipher_3Way, '', '');
    RegisterCipher(TCipher_Blowfish, '', '');
    RegisterCipher(TCipher_Cast128, 'Cast 128', '');
    RegisterCipher(TCipher_Cast256, 'Cast 256', 'Patented');
    RegisterCipher(TCipher_FROG, '', '');
    RegisterCipher(TCipher_Gost, '', '');
    RegisterCipher(TCipher_IDEA, '', 'free for non-commercial');
    RegisterCipher(TCipher_Mars, '', 'Patented');
    RegisterCipher(TCipher_Misty, 'Misty 1', 'free for non-commercial');
    RegisterCipher(TCipher_Q128, '', '');
    RegisterCipher(TCipher_RC2, '', '');
    RegisterCipher(TCipher_RC4, '', 'Patented by RSA');
    RegisterCipher(TCipher_RC5, '', 'Patented by RSA');
    RegisterCipher(TCipher_RC6, '', 'Patented by RSA');
    RegisterCipher(TCipher_Rijndael, '', '');
    RegisterCipher(TCipher_SAFER_K40, 'SAFER-K40', '');
    RegisterCipher(TCipher_SAFER_SK40, 'SAFER-SK40', 'Keyscheduling');
    RegisterCipher(TCipher_SAFER_K64, 'SAFER-K64', '');
    RegisterCipher(TCipher_SAFER_SK64, 'SAFER-SK64', 'Keyscheduling');
    RegisterCipher(TCipher_SAFER_K128, 'SAFER-K128', '');
    RegisterCipher(TCipher_SAFER_SK128, 'SAFER-SK128', 'Keyscheduling');
    RegisterCipher(TCipher_SCOP, '', '');
    RegisterCipher(TCipher_Shark, '', '');
    RegisterCipher(TCipher_Skipjack, '', '');
    RegisterCipher(TCipher_TEA, 'TEA', '');
    RegisterCipher(TCipher_TEAN, 'TEA extended', '');
    RegisterCipher(TCipher_Twofish, '', '');
  end;
finalization
  FHashList.Free;
  FHashList := nil;
  FCipherList.Free;
  FCipherList := nil;
end.
