unit XWModExpt;
{: Defines the base expert class for Xapware Modules. }
interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  xwViewFrm, Menus, ExtCtrls, ComCtrls;
const
  { Module registration location in the Windows Registry. }
  cModRegLocation = 'Software\Xapware Technologies Inc\Modules';
  cModRegSection  = 'InitModules';  // Module initialization data section
  cModuleMngrGroupIndex     = 1; // Group index for module manager menu for menu merging

type

  // forward declaration
  TXWModuleInterface = class;

  TXWModuleInterfaceClass = class of TXWModuleInterface;

  TXWModuleManager = class(TObject)
  private
    FCurrentModule: TXWModuleInterface; // Current instance
    FCurrentModuleHandle: HMODULE;      // Actually handle to the package
    FModuleList: TStringList;           // List of available modules
    FActiveMenuRef: TMenuItem;          // Reference to active module menu
    FCloseModuleMenuIndex: Integer;     // Reference to menu for closing the module

    FModuleManagerMenu: TMainMenu;      // Module Manager level menu

    FViewParentWindow: TWinControl;     // View parent window

    FModuleLevelMenuParent: TMenuItem;  // Module level parent menu
    FModuleLevelTBParent: TPanel;       // Module level toolbar parent

    FViewMenu: TMenuItem;               // View level menu
    FViewTBParent: TPanel;              // View toolbar parent


    // Events
    FBeforeLoadModule: TNotifyEvent;
    FAfterLoadModule: TNotifyEvent;
    FBeforeUnloadModule: TNotifyEvent;
    FAfterUnloadModule: TNotifyEvent;

    procedure SetViewParentWindow(const Value: TWinControl);

  protected
    procedure ModuleMenuOnClick(Sender: TObject);
    procedure CloseModuleOnClick(Sender: TObject);

  public
    constructor Create;
    destructor Destroy; override;

    // Load the module specified by AModuleName.
    procedure LoadModule(const AModuleName: String);
    // Unload the currently loaded module.
    procedure UnloadModule;

    // Open the view (form) specified by AIndex in the current modules view list.
    function OpenView(const AIndex: Integer; AWithParent: Boolean): TViewForm;
    // Close the currently loaded view.
    procedure CloseView;

    // List of registered modules.
    property ModuleList: TStringList read FModuleList;
    // Reference to the current module (package)
    property CurrentModule: TXWModuleInterface read FCurrentModule;
    // Reference to the current module handle
    property CurrentModuleHandle: HMODULE read FCurrentModuleHandle;
    // Menu for the module manager that gets merged with the Shell Applications's menu
    property ModuleManagerMenu: TMainMenu read FModuleManagerMenu;
    // Parent window for each view form. This is a panel on the Shell App's main form.
    property ViewParentWindow: TWinControl read FViewParentWindow write SetViewParentWindow;
    // Module specific menu items - merged with Shell App's main menu.
    property ModuleLevelMenuParent: TMenuItem read FModuleLevelMenuParent write FModuleLevelMenuParent;
    // Module specific tool bar - merged with Shell App's Coolbar
    property ModuleLevelTBParent: TPanel read FModuleLevelTBParent write FModuleLevelTBParent;
    // View specific menu items - merged with Shell App's main menu.
    property ViewMenu: TMenuItem read FViewMenu write FViewMenu;
    // View specific tool bar - merged with Shell App's Coolbar
    property ViewTBParent: TPanel read FViewTBParent write FViewTBParent;

    { Events for which the Shell App can provide Event handlers }
    property BeforeLoadModule: TNotifyEvent read FBeforeLoadModule write FBeforeLoadModule;
    property AfterLoadModule:  TNotifyEvent read FAfterLoadModule write FAfterLoadModule;
    property BeforeUnloadModule: TNotifyEvent read FBeforeUnloadModule write FBeforeUnloadModule;
    property AfterUnloadModule:  TNotifyEvent read FAfterUnloadModule write FAfterUnloadModule;

  end;


  {: The TXWModuleInterface class defines the Builder class for constructing concrete
     Modules. It contains the logic for providing references to modules and thier views
     (Forms). }
  TXWModuleInterface = class(TPersistent)
  protected
    FCurrentView: TViewForm;            // Currently loaded view form.

    FModuleLevelMenu: TPopupMenu;       // Module specific menu
    FViewLevelMenu: TMenuItem;          // View specific menu

    FModuleLevelMenuParent: TMenuItem;  // Parent menu to the module level menu items.
    FModuleLevelTBParent: TPanel;       // Parent panel for module toolbars

    FViewParentWindow: TWinControl;    // Parent window to views (Forms) }
    FViewLevelTBParent: TPanel;        // Parent panel for view toolbars


    {: Override GetModuleCaption to return a Caption to be used in the ShellApp }
    function GetModuleCaption: String; virtual; abstract;
    {: Override GetModuleDescription to return a description of this module }
    function GetModuleDescription: String; virtual; abstract;
    {: Override GetModuleFileName to return a Module's file name }
    function GetModuleFileName: String; virtual; abstract;
    {: Override GetModuleIcon to return the Module specific icon }
    function GetModuleIcon: TIcon; virtual; abstract;
    {: Override GetModuleLevelMenu to return a module level Popup menu }

    function GetModuleLevelMenu: TPopupMenu; virtual;
    {: Override GetModuleToolbar to return a module specific image list. }
    function GetModuleToolBar: TToolBar; virtual; abstract;

    // View methods

    {: Override GetView to retrieve the view of the given index and parent set up
       accordingly. }
    function GetView(const AIndex: Integer; AWithParent: Boolean): TViewForm; virtual; abstract;
    {: Override CloseView to close the currently open view. }
    procedure CloseView; virtual; abstract;
    {: Override GetViewCount to return the number of views contained by this module. }
    function GetViewCount: Integer; virtual; abstract;
    {: Override GetViewDescriptions to return the list of descriptions for each view }
    function GetViewDescriptions: TStringList; virtual; abstract;
    {: Override GetViewCaptions to return a list of captions for each view }
    function GetViewCaptions: TStringList; virtual; abstract;
    {: Override GetCurrentView to return the current view form }
    function GetCurrentView: TViewForm; virtual; abstract;

    {: Methods for loading/unloading module (Package) level menu items and toolbars.}
    procedure LoadModuleLevelMenu;
    procedure UnloadModuleLevelMenu;
    procedure LoadModuleLevelToolbar;

    {: Methods for loading/unloading view (Form) level menu item and toolbars. }
    procedure LoadViewLevelMenu;
    procedure UnloadViewLevelMenu;
    procedure LoadViewLevelToolBar;

  public
    constructor Create(AParentWindow: TWinControl; AModuleLevelMenuParent: TMenuItem;
      AModuleLevelTBParent: TPanel); virtual;
    destructor Destroy; override;
    {: OpenView is a Template Method that opens the view and calls the appropriate
       methods for menu/toolbar merging. }
    function OpenView(const AIndex: Integer; AWithParent: Boolean;
      AViewMenu: TMenuItem; AViewToolBarParent: TPanel): TViewForm; virtual;
   {: Override CanSwitchModules to indicate weather or not the module can be switched.
      You may want to prevent the user from switching modules such as when a record on
      a view is in insert or edit mode. }
    function CanSwitchModules: Boolean; virtual;
    {: Override ModuleLogOn to handle the login logic for database connections module }
    function ModuleLogOn(const AUserName, APassword: String;
       const LenderID: Integer): Boolean; virtual;

    // Module Properties
    property ModuleCaption: String read GetModuleCaption;
    property ModuleDescripton: String read GetModuleDescription;
    property ModuleFileName: String read GetModuleFileName;
    property ModuleIcon: TIcon read GetModuleIcon;
    property ModuleToolBar: TToolBar read GetModuleToolBar;

    // View Properties
    property ViewCount: Integer read GetViewCount;
    property ViewDescriptions: TStringList read GetViewDescriptions;
    property ViewCaptions: TStringList read GetViewCaptions;
    property CurrentView: TViewForm read GetCurrentView;
  published
  end;

implementation
uses Registry;

{ Helper functions }

function RemoveExt(const AFileName: String): String;
{ Removes the extension from a file name. }
begin
  if Pos('.', AFileName) <> 0 then
    Result := Copy(AFileName, 1, Pos('.', AFileName)-1)
  else
    Result := AFileName;
end;

function GetModuleClassName(const AModuleName: String): String;
{ The Actual class name of the TXWModuleInterface is provided in the registry. This
  method retrieves that class name. }
var
  IniFile: TRegIniFile;
begin
  IniFile := TRegIniFile.Create(cModRegLocation);
  try
    Result := IniFile.ReadString(RemoveExt(AModuleName), 'ClassName', EmptyStr);
  finally
    IniFile.Free;
  end;
end;


{ TXWModuleManager }

function TXWModuleManager.OpenView(const AIndex: Integer;
  AWithParent: Boolean): TViewForm;
{ This method opens a view specified by the view index, AIndex. The parent TMenuItem
  and parent container for the view toolbar are passed to the Module Interface OpenView()
  method in order to allow these items to be merged with the Shell Apps menu and
  toolbar. }
begin
  if CurrentModule.CurrentView <> nil then
    CloseView;
  Result := CurrentModule.OpenView(AIndex, AWithParent, ViewMenu, ViewTBParent);
  Result.Show;
end;

procedure TXWModuleManager.CloseView;
begin
  CurrentModule.UnloadViewLevelMenu;
  CurrentModule.CloseView;
end;

procedure TXWModuleManager.LoadModule(const AModuleName: String);
var
  ModClassName: String;
  XWMIC: TXWModuleInterfaceClass;

begin

  // Call any event handlers provided by the client.
  if Assigned(FBeforeLoadModule) then
    FBeforeLoadModule(self);

  // Unload any previously loaded modules.
  if FCurrentModuleHandle <> 0 then
    UnloadModule;

  try
    // Load the specified module
    FCurrentModuleHandle := LoadPackage(AModuleName);

    // Return the classname that needs to be instanciated into ModClassName
    ModClassName := GetModuleClassName(AModuleName);

    { Create an instance of the class using the FindClass() procedure. Note, that this
      requires that the class already be registered with the streaming system using
      RegisterClass().  This is done in a unit initialization section in each module
      package. }
    XWMIC := TXWModuleInterfaceClass(FindClass(ModClassName));
    FCurrentModule := XWMIC.Create(FViewParentWindow, FModuleLevelMenuParent,
      FModuleLevelTBParent);

    if Assigned(FAfterLoadModule) then
      FAfterLoadModule(self);

    { Load the module's menus and toolbars. }
    FCurrentModule.LoadModuleLevelMenu;
    FCurrentModule.LoadModuleLevelToolbar;

  except
    on E: Exception do
    begin
      UnloadModule;
      raise;
    end;
  end;
end;

procedure TXWModuleManager.UnloadModule;
begin

  if FCurrentModuleHandle <> 0 then
  begin

    if Assigned(FBeforeUnloadModule) then
      FBeforeUnloadModule(self);

    { Unloaded any module level and view level menus before unloading and freeing this
      module. }
    if FCurrentModule <> nil then
    begin
      FCurrentModule.UnloadModuleLevelMenu;
      CloseView;
      FCurrentModule.Free;
      FCurrentModule := nil;
    end;

    // Unregister any classes provided by the module
    UnRegisterModuleClasses(FCurrentModuleHandle);
    // Unload the module package
    UnloadPackage(FCurrentModuleHandle);

    FCurrentModuleHandle := 0;

    // Call any event handlers provided by the client
    if Assigned(FAfterUnloadModule) then
      FAfterUnloadModule(self);
  end;
end;

destructor TXWModuleManager.Destroy;
begin
  FModuleList.Free;
  UnloadModule;
  FModuleManagerMenu.Free;
  inherited;
end;

constructor TXWModuleManager.Create;
{ All available modules are registered in the Windows Registry.  When the Module Manager
  is created, it loads this information and creates menu items on the Shell Apps
  main menu. When the user selects one of these menu items, the module for which the
  menu was created is loaded. }
var
  IniFile: TRegIniFile;
  MenuItem: TMenuItem;
  i: integer;
begin
  inherited;
  FCurrentModuleHandle := 0;
  FModuleManagerMenu := TMainMenu.Create(nil);

  FModuleList := TStringList.Create;

  IniFile := TRegIniFile.Create(cModRegLocation);
  try
    IniFile.ReadSectionValues(cModRegSection, FModuleList);
  finally
    IniFile.Free;
  end;

  // Add the Sections to the module main menu
  MenuItem := TMenuItem.Create(FModuleManagerMenu);
  MenuItem.Caption := '&Xapware Modules';
  MenuItem.GroupIndex := cModuleMngrGroupIndex;
  FModuleManagerMenu.Items.Add(MenuItem);

  // Add Menu items for each module.
  for i := 0 to FModuleList.Count - 1 do
  begin
    MenuItem := TMenuItem.Create(FModuleManagerMenu);
    MenuItem.Caption := FModuleList.Names[i];
    MenuItem.OnClick := ModuleMenuOnClick;
    FModuleManagerMenu.Items[0].Add(MenuItem);
  end;

  // Create Separator
  MenuItem := TMenuItem.Create(FModuleManagerMenu);
  MenuItem.Caption := '-';
  FModuleManagerMenu.Items[0].Add(MenuItem);

  // Create Close Module menu item
  MenuItem := TMenuItem.Create(FModuleManagerMenu);
  MenuItem.Caption := '&Close Module';
  MenuItem.OnClick := CloseModuleOnClick;
  MenuItem.Enabled := False;
  FModuleManagerMenu.Items[0].Add(MenuItem);

  { Save a reference to the index of the menu item required to close a module. This
    will be referred to in other methods. }
  FCloseModuleMenuIndex := MenuItem.MenuIndex;

end;

procedure TXWModuleManager.SetViewParentWindow(const Value: TWinControl);
begin
  FViewParentWindow := Value;
end;

procedure TXWModuleManager.ModuleMenuOnClick(Sender: TObject);
{ This method is the default event handler for the Module Manager's menu items. These
  menu items refer to the list of available modules so when the selects one, it's
  cooresponding module is loaded. } 
var
  Name: String;
  ModFileName: String;
begin
  // The Caption represents the module name.
  Name := (Sender as TMenuItem).Caption;
  // Get the actual Package file name.
  ModFileName := FModuleList.Values[Name];

  // Unload the previously loaded package
  if FCurrentModule <> nil then
    UnloadModule;

  // Load the new package.
  LoadModule(ModFileName);

  // Disable the menu item to load this module.
  FModuleManagerMenu.Items[0].Items[FCloseModuleMenuIndex].Enabled :=
    FCurrentModule <> nil;

  { Uncheck the previously checked menu item from the last module. Check the new
    module's menu item. }

  if FCurrentModule <> nil then
  begin
    if FActiveMenuRef <> nil then
      FActiveMenuRef.Checked := False;
    FActiveMenuRef := TMenuItem(Sender);
    FActiveMenuRef.Checked := True;
  end;

end;

procedure TXWModuleManager.CloseModuleOnClick(Sender: TObject);
begin
{ Unload the module and uncheck its menu item. }
  UnloadModule;
  FModuleManagerMenu.Items[0].Items[FModuleManagerMenu.Items[0].Count - 1].Enabled :=
    FCurrentModule <> nil;
  if FCurrentModule = nil then
    if FActiveMenuRef <> nil then
    begin
      FActiveMenuRef.Checked := False;
      FActiveMenuRef := nil;
    end;
end;

!@#$

{ TXWModuleExpert }

function TXWModuleInterface.CanSwitchModules: Boolean;
begin
  Result := True; // Default
end;

constructor TXWModuleInterface.Create(AParentWindow: TWinControl;
  AModuleLevelMenuParent: TMenuItem; AModuleLevelTBParent: TPanel);
{ This constructor sets up the reference to the parent window that will contain the
  veiw forms. It also sets up references for the Shell Apps main menu and containers
  for the module level menu items and toolbars. }
begin
  inherited Create;
  FViewParentWindow := AParentWindow;
  FModuleLevelMenu := nil;
  FModuleLevelMenuParent := AModuleLevelMenuParent;
  FModuleLevelTBParent   := AModuleLevelTBParent;
end;

destructor TXWModuleInterface.Destroy;
begin
  inherited Destroy;
end;

function TXWModuleInterface.GetModuleLevelMenu: TPopupMenu;
begin
  Result := nil; // Default
end;

procedure TXWModuleInterface.LoadViewLevelMenu;
var
  MenuItem: TMenuItem;
  i: integer;
begin
  { Since there is only one level of menu merging, we have to manually create menu items
    for both the Module and View level menus. }
  if FCurrentView <> nil then
  begin
    if FCurrentView.ViewMenu <> nil then
    begin
      for i := 0 to FCurrentView.ViewMenu.Items.Count - 1 do
      begin
        MenuItem := TMenuItem.Create(FViewLevelMenu);
        MenuItem.Caption := FCurrentView.ViewMenu.Items[i].Caption;
        MenuItem.OnClick := FCurrentView.ViewMenu.Items[i].OnClick;
        FViewLevelMenu.Add(MenuItem);
      end;
      FViewLevelMenu.Visible := True;
    end;
  end;
end;

procedure TXWModuleInterface.UnloadViewLevelMenu;
var
  i: integer;
begin
  if FViewLevelMenu <> nil then
  begin
    // Free the view level menus
    for i := FViewLevelMenu.Count - 1 downto 0 do
      FViewLevelMenu.Delete(i);

    FViewLevelMenu.Visible := False;
  end;
end;

procedure TXWModuleInterface.LoadViewLevelToolBar;
begin
  if FCurrentView <> nil then
    if FCurrentView.ViewToolbar <> nil then
    begin
      FCurrentView.ViewToolBar.EdgeBorders := [];
      FCurrentView.ViewToolBar.Parent := FViewLevelTBParent;
    end;
end;

function TXWModuleInterface.ModuleLogOn(const AUserName, APassword: String;
  const LenderID: Integer): Boolean;
begin
  Result := True; // Default
end;

function TXWModuleInterface.OpenView(const AIndex: Integer; AWithParent: Boolean;
AViewMenu: TMenuItem; AViewToolBarParent: TPanel): TViewForm;
begin
  { Notice the use of this Template Method to set up the menu/toolbar merging items and
    then to call the abstract function GetView().  Each implementation of
    TWModuleInterface must implement the GetView method to return a valid TViewForm }
  FViewLevelMenu := AViewMenu;
  FViewLevelTBParent := AViewToolbarParent;

  Result := GetView(AIndex, AWithParent);

  { Everytime a view is loaded, call the methods to merge its menu with that specified
    by FViewMenu as well as the toolbar specified by FViewTBParent.}
  LoadViewLevelMenu;
  LoadViewLevelToolBar;
end;

procedure TXWModuleInterface.LoadModuleLevelMenu;
var
  MenuItem: TMenuItem;
  i: integer;
begin
  if FModuleLevelMenu <> nil then
  begin
    for i := 0 to FModuleLevelMenu.Items.Count - 1 do
    begin
      MenuItem := TMenuItem.Create(FModuleLevelMenuParent);
      MenuItem.Caption := FModuleLevelMenu.Items[i].Caption;
      MenuItem.OnClick := FModuleLevelMenu.Items[i].OnClick;
      FModuleLevelMenuParent.Add(MenuItem);
    end;
    FModuleLevelMenuParent.Visible := True;
  end;
end;

procedure TXWModuleInterface.UnloadModuleLevelMenu;
var
  i: integer;
begin
  for i := FModuleLevelMenuParent.Count - 1 downto 0 do
    FModuleLevelMenuParent.Delete(i);

  FModuleLevelMenuParent.Visible := False;
end;

procedure TXWModuleInterface.LoadModuleLevelToolbar;
begin
  if ModuleToolBar <> nil then
  begin
     ModuleToolBar.EdgeBorders := [];
     ModuleToolBar.Parent := FModuleLevelTBParent;
  end;
end;

end.

