unit kbmMemTable;

// TKbmMemTable v. 1.09
// =========================================================================
// An inmemory temporary table.
// Can be used as a demonstration of how to create descendents of TDataSet,
// or as in my case, to allow a program to generate temporary data that can
// be used directly by all data aware controls.
//
// Copyright 1999 Kim Bo Madsen/Optical Services - Scandinavia
// All rights reserved.
//
// You are allowed to used this component in any project for free.
// You are NOT allowed to claim that you have created this component or to
// copy its code into your own component and claim that is was your idea.
// Im offering this for free for your convinience, and the ONLY thing I request
// is to get an e-mail about what project this component (or derived versions)
// is used for. That will be my reward of offering this component for free to you!
//
// You dont need to state my name in your software, although it would be
// appreciated if you do.
//
// If you find bugs or alter the component (f.ex. see suggested enhancements
// further down), please DONT just send the corrected/new code out on the internet,
// but instead send it to me, so I can put it into the official version. You will
// be acredited if you do so.
//
//
// DISCLAIMER
// By using this component or parts theirof you are accepting the full
// responsibility of the use. You are understanding that the author cant be
// made responsible in any way for any problems occuring using this component.
// You also recognize the author as the creator of this component and agrees
// not to claim otherwize!
//
// Please forward corrected versions (source code ONLY!), comments,
// and emails saying you are using it for this or that project to:
//            kbm@optical.dk
//
// Suggestions for future enhancements:
//
//      - Index handling functions.
//      - Sorting.
//      - Support for SetKey, FindKey, FindNearest
//      - Support for MasterField, Lookupfields.
//      - IDE designer for adding static data to the memtable.
//
// History:
//
//1.00:	The first release. Was created due to a need for a component like this.
//                                                                    (15. Jan. 99)
//1.01:	The first update. Release 1.00 contained some bugs related to the ordering
//	of records inserted and to bookmarks. Problems fixed.         (21. Jan. 99)

//1.02:	Fixed handling of NULL values. Added SaveToStream, SaveToFile,
//	LoadFromStream and LoadFromFile. SaveToStream and SaveToFile is controlled
//	by a flag telling if to save data, contents of calculated fields,
//	contents of lookupfields and contents of non visible fields.
//	Added an example application with Delphi 3 source code.       (26. Jan. 99)
//
//1.03: Claude Rieth from Computer Team sarl (clrieth@team.lu) came up with an
//      implementation of CommaText and made a validation check in _InternalInsert.
//      Because I allready have implemented the saveto.... functions, I decided
//      to implement Claude's idea using my own saveto.... functions. (27. Jan. 99)
//      I have decided to rename the component, because Claude let me know that
//      the RX library have a component with the same name as this.
//      Thus in the future the component will be named TkbmMemTable.
//      SaveToStream and LoadFromStream now set up date and decimal separator
//      temporary to make sure that the data saved can be loaded on another
//      installation with different date and decimal separator setups.
//      Added EmptyTable method to clear the contents of the memory table.
//
//1.04: Wagner ADP (wagner@cads-informatica.com.br) found a bug in the _internalinsert
//      procedure which he came up with a fix for.                     (4. Feb. 99)
//      Added support for the TDataset protected function findrecord.
//      Added support for CreateTable, DeleteTable.
//
//1.05: Charlie McKeegan from Task Software Limited (charlie@task.co.uk) found
//      a minor bug (and came up with a fix) in SetRecNo which mostly is
//      noticeable in a grid where the grid wont refresh when the record slider
//      is moved to the end of the recordset.                          (5. Feb. 99)
//      Changed SaveToStream to use the displayname of the field as the header
//      instead of the fieldname.
//1.06: Introduced a persistence switch and a reference to a file. If the
//      persistence switch is set, the file will be read aut. on table open,
//      and the contents of the table will be saved in the table on table close.
//1.07: Changed calculation of fieldofsets in InternalOpen to follow the fielddefs
//      instead of fields. It has importance when rearranging fields.
//      Because of this change the calculation of fieldoffsets into the buffer
//      has been changed too. These corrections was found to be needed to
//      support the new components tkbmPooledQuery and tkbmPooledStoredProc
//      (found in the package tkbmPooledConn)
//      which in turn handles pooled connections to a database. Very usefull
//      in f.ex a WWW application as they also makes the BDE function threadsafe,
//      and limits the concurrent connections to a database.
//1.08: Changed buffer algorithm in GetLine since the old one was faulty.
//      Problem was pointed out by: Markus Roessler@gmx.de
//1.09: Added LoadFromDataset, SaveToDataset, CreateTableAs,
//      BCD and BLOB support by me.
//      James Baile (James@orchiddata.demon.co.uk) pointed out a bug in GetRecNo
//      which could lead to an endless loop in cases with filtering. He also
//      provided code for sorting, which I have been rearranging a bit and
//      implemented.
//      Travis Diamond (tdiamond@airmail.net) pointed out a bug in GetWord where
//      null fields would be skipped.
//=============================================================================

{$ifndef VER100} // C++ Builder 3
{$ObjExportAll On}
{$ASSERTIONS ON}
{$endif}

interface

uses SysUtils,Classes,Db;

type
    EMemTableError = class(Exception);

    TRecInfo=record
        Bookmark: longint;
        RecordNo: integer;
        BookmarkFlag: TBookmarkFlag;
    end;
    PRecInfo=^TRecInfo;

{
Internal buffer layout:
+------------------------+------------------------+---------------------------+
|     RECORD DATA        |    Rec.Information     |     Calculated Fields     |
| Record length bytes    |  SizeOf(TRecInfo) bytes|    CalcFieldSize bytes    |
+------------------------+------------------------+---------------------------+
                         ^                        ^
                    StartRecInfo              StartCalculated

Blobsfields in the internal buffer are pointers to the blob data.
}

  TkbmMemTableSaveFlag = (mtfSaveData, mtfSaveCalculated, mtfSaveLookup,mtfSaveNonVisible);
  TkbmMemTableSaveFlags = set of TkbmMemTableSaveFlag;

  TkbmBlob = class
  private
     FLength:cardinal;
     FBuffer:PChar;
  public
     constructor Create;
     destructor Destroy; override;

     procedure LoadFromStream(stream:TMemoryStream);
     procedure SaveToStream(stream:TMemoryStream);
     property Length:cardinal read FLength write FLength;
     property Buffer:Pchar read FBuffer write FBuffer;
  end;
  PkbmBlob=^TkbmBlob;

  TkbmMemTable = class(TDataSet)
  private
        FIsOpen:                                Boolean;
        FRecNo:                                 integer;
        FFilterBuffer:                          PChar;
        FRecords:                               TList;
        FBufferSize,
        FStartRecInfo,
        FStartCalculated:integer;
        FRecordSize:                            integer;
        FFieldOfs:                              array [0..255] of integer;
        FReadOnly:                              boolean;
        FPersistent:                            boolean;
        FPersistentFile:                        string;
        FBlobs:                                 boolean;
        FIndexList:                             TList;
        FCaseInsensitiveSort:                   boolean;
        FDescendingSort:                        boolean;
        FDummyStr:                              string;
        FIndexFieldNames:                       string;
        function GetActiveRecordBuffer:         PChar;
        function FilterRecord(Buffer: PChar):   Boolean;
        procedure _InternalAdd(Buffer:Pointer);
        procedure _InternalDelete(Pos:integer);
        procedure _InternalInsert(Pos:integer; Buffer:Pointer);
        procedure _InternalEmpty;
        procedure _InternalFirst;
        procedure _InternalLast;
        function  _InternalNext:boolean;
        function  _InternalPrior:boolean;
        procedure _InternalFreeRecord(RecPtr:Pointer);
        function _InternalMoveRecord(Source, Destination: Integer): Boolean;

        procedure BuildFieldList(List:TList; const FieldNames: string);
        procedure QuickSort(L, R: Integer);
        function  CompareRecords(FieldList:TList; Record1,Record2:PChar): Integer;
  protected
        procedure InternalOpen; override;
        procedure InternalClose; override;
        procedure InternalFirst;override;
        procedure InternalLast;override;

        procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
        procedure InternalDelete; override;
        procedure InternalInitRecord(Buffer: PChar); override;
        procedure InternalPost; override;

        procedure InternalInitFieldDefs; override;
        procedure InternalSetToRecord(Buffer: PChar); override;

        procedure DoBeforeClose; override;
        procedure DoAfterOpen; override;

        function IsCursorOpen: Boolean; override;
        function GetCanModify: Boolean; override;
        function GetRecordSize: Word;override;
        function GetRecordCount: integer;override;

        function AllocRecordBuffer: PChar; override;
        procedure FreeRecordBuffer(var Buffer: PChar); override;

        function CalcFieldSize(FieldType:TFieldType; Size:longint):longint;
        function GetFieldPointer(Buffer:PChar; Field:TField):PChar;

        procedure SetFieldData(Field: TField; Buffer: Pointer);override;
{$IFNDEF VER125}  // C++ Builder 4
 {$IFNDEF VER120} // Delphi 4
        function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
 {$ENDIF}
{$ENDIF}

        function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
        function FindRecord(Restart, GoForward: Boolean): Boolean; override;

        function GetRecNo: integer;override;
        procedure SetRecNo(Value: integer);override;

        function GetIsIndexField(Field: TField): Boolean; override;

        function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
        procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
        procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
        procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
        procedure InternalGotoBookmark(Bookmark: Pointer); override;

        function BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean; override;
        function CurrToBCD(const Curr: Currency; BCD: Pointer; Precision, Decimals: Integer): Boolean; override;

        procedure InternalHandleException; override;

        procedure SetCommaText(AString: String);
        function GetCommaText: String;
        function CopyRecords(Source,Destination:TDataSet; Count:longint):longint;
        function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;

        function MoveCurRec(Destination:Longint):Boolean;
  public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
{$IFDEF VER125} // C++ Builder 4
        function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
{$ENDIF}
{$IFDEF VER120} // Delphi 4
        function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
{$ENDIF}
        procedure CreateTable;
        procedure CreateTableAs(Source:TDataSet);
        procedure DeleteTable;
        procedure LoadFromFile(const FileName: string);
        procedure LoadFromStream(Stream: TStream);
        procedure LoadFromDataSet(Source:TDataSet);
        procedure SaveToFile(const FileName: string; flags:TkbmMemTableSaveFlags);
        procedure SaveToStream(Stream: TStream; flags:TkbmMemTableSaveFlags);
        procedure SaveToDataSet(Destination:TDataSet);
        procedure EmptyTable;
        property CommaText:string read GetCommaText write SetCommaText;
        procedure Sort;
        function IsSequenced:Boolean; override;
  published
        property Active;
        property Filtered;
        property IndexFields:string read FIndexFieldNames write FIndexFieldNames;
        property CaseInsensitiveSort:boolean read FCaseInsensitiveSort write FCaseInsensitiveSort;
        property DescendingSort:boolean read FDescendingSort write FDescendingSort;
        property ReadOnly:boolean read FReadOnly write FReadOnly default false;
        property PersistentFile:string read FPersistentFile write FPersistentFile;
        property Persistent:boolean read FPersistent write FPersistent default false;
        property Filter:string read FDummyStr;
        property BeforeOpen;
        property AfterOpen;
        property BeforeClose;
        property AfterClose;
        property BeforeInsert;
        property AfterInsert;
        property BeforeEdit;
        property AfterEdit;
        property BeforePost;
        property AfterPost;
        property BeforeCancel;
        property AfterCancel;
        property BeforeDelete;
        property AfterDelete;
        property BeforeScroll;
        property AfterScroll;
        property OnCalcFields;
        property OnDeleteError;
        property OnEditError;
        property OnFilterRecord;
        property OnNewRecord;
        property OnPostError;
  end;

  TkbmBlobStream = class(TMemoryStream)
  private
    FField: TBlobField;
    FDataSet: TkbmMemTable;
    FMode:TBlobStreamMode;
    FPBlob: PkbmBlob;
    FFieldNo: Integer;
    FModified: Boolean;
    procedure ReadBlobData;
    procedure WriteBlobData;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    procedure Truncate;
  end;

procedure Register;

implementation

uses
  TypInfo, Dialogs, Windows, Forms, Controls, IniFiles;

const
  TkbmSupportedFieldTypes=[ftString,ftSmallint,ftInteger,ftWord,ftBoolean,ftFloat,ftCurrency,
                          ftDate,ftTime,ftDateTime,ftAutoInc,ftBCD,ftBlob,ftMemo,ftGraphic,ftFmtMemo,
                          ftParadoxOle,ftDBaseOle,ftTypedBinary];
  TkbmBlobTypes=[ftBlob,ftMemo,ftGraphic,ftFmtMemo,ftParadoxOle,ftDBaseOle,ftTypedBinary];

// Compare two fields.
function CompareFields(p1,p2:pointer; FieldType: TFieldType; CaseInsensitive: Boolean):Integer;
begin
     Result := 0;
     case FieldType of
       ftString:
          if CaseInsensitive then
             Result:=AnsiCompareText(PChar(p1), PChar(p2))
          else
              Result:=AnsiCompareStr(PChar(p1), PChar(p2));

       ftSmallint:
          Result:=SmallInt(p1^)-SmallInt(p2^);

       ftInteger,
       ftDate,
       ftTime,
       ftAutoInc:
          Result:=Longint(p1^)-Longint(p2^);

       ftWord:
          Result:=Word(p1^)-Word(p2^);

       ftBoolean:
          if WordBool(p1^)>WordBool(p2^) then Result:=1
          else if WordBool(p1^)<WordBool(p2^) then Result:=-1;

       ftFloat,
       ftCurrency:
          if Double(p1^)>Double(p2^) then Result:=1
          else if Double(p1^)<Double(p2^) then Result:=-1;

       ftDateTime:
          if TDateTime(p1^)>TDateTime(p2^) then Result:=1
          else if TDateTime(p1^)<TDateTime(p2^) then Result:=-1;
     end;
end;

// Compare two records.
function TkbmMemTable.CompareRecords(FieldList:TList; Record1,Record2:PChar): Integer;
var
   p1,p2:PChar;
   fld:TField;
   i:integer;
begin
     Result:=0;

     // Loop through all indexfields, left to right.
     for i:=0 to FieldList.Count-1 do
     begin
          fld:=TField(FieldList[i]);

          // Get data for specified field for the two records.
          p1:=GetFieldPointer(Record1,fld);
          p2:=GetFieldPointer(Record2,fld);

          // Check if both not null.
          if Boolean(p1[0]) and Boolean(p2[0]) then
          begin
               // Skip null flag.
               inc(p1);
               inc(p2);

               // Compare the fields.
               Result:=CompareFields(p1,p2,fld.DataType,FCaseInsensitiveSort);
          end
          else if Boolean(p1[0]) then Result:=1
          else if Boolean(p2[0]) then Result:=-1;
          if Result<>0 then break;
     end;

     // Couldnt sort them according to fieldcontents, will now sort according to recnum.
     if Result=0 then
        Result:=PRecInfo(Record1+FStartRecInfo)^.RecordNo - PRecInfo(Record2+FStartRecInfo)^.RecordNo;

     // If descending sort, invert result.
     if FDescendingSort then Result:=-Result;
end;

constructor TkbmMemTable.Create(AOwner: TComponent);
begin
     inherited Create(AOwner);
     FRecords:=TList.Create;
     FPersistent:=false;
     FIndexList:=TList.Create
end;

destructor TkbmMemTable.Destroy;
begin
     _InternalEmpty;
     inherited Destroy;

     // Delete allocated records.
     FIndexList.free;
     FIndexList:=nil;
     FRecords.free;
     FRecords:=nil;
end;

function TkbmMemTable.CalcFieldSize(FieldType:TFieldType; Size:longint):longint;
begin
     case FieldType of
          ftString:             Result:=Size+1;
          ftSmallInt:           Result:=SizeOf(SmallInt);
          ftInteger:            Result:=SizeOf(Integer);
          ftWord:               Result:=SizeOf(Word);
          ftBoolean:            Result:=SizeOf(WordBool);
          ftFloat:              Result:=SizeOf(Double);
          ftCurrency:           Result:=SizeOf(Double);
          ftDate:               Result:=SizeOf(Integer);
          ftTime:               Result:=SizeOf(Integer);
          ftDateTime:           Result:=SizeOf(Double);
          ftAutoInc:            Result:=SizeOf(Integer);
          ftBlob:               Result:=SizeOf(PkbmBlob);
          ftMemo:               Result:=SizeOf(PkbmBlob);
          ftGraphic:            Result:=SizeOf(PkbmBlob);
          ftFmtMemo:            Result:=SizeOf(PkbmBlob);
          ftParadoxOle:         Result:=SizeOf(PkbmBlob);
          ftDBaseOle:           Result:=SizeOf(PkbmBlob);
          ftTypedBinary:        Result:=SizeOf(PkbmBlob);
          ftBCD:                Result:=34;
     else
          Result:=0;
     end;
end;

function TkbmMemTable.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
     Result:=TkbmBlobStream.Create(Field as TBlobField, Mode);
end;

procedure TkbmMemTable.CreateTable;
var
   i:Integer;
begin
     CheckInactive;

     // If no fielddefs existing, use the previously defined fields.
     if FieldDefs.Count = 0 then
        for i:=0 to FieldCount-1 do
            with Fields[i] do
                 if FieldKind = fkData then
                    FieldDefs.Add(FieldName, DataType, Size, Required);

     // Remove previously defined fields and create new from fielddefs.
     DestroyFields;
     CreateFields;
end;

// Create memory table as another dataset.
procedure TkbmMemTable.CreateTableAs(Source:TDataSet);
var
   i:integer;
begin
     CheckInactive;

     if Source=nil then exit;

     // Add fields as they are defined in the other dataset.
     Source.FieldDefs.Update;
     FieldDefs.Assign(Source.FieldDefs);

     // Remove non supported fieldsdefs.
     for i:=FieldDefs.Count-1 downto 0 do
         if not (FieldDefs.Items[i].DataType in TkbmSupportedFieldTypes) then
            FieldDefs.Items[i].free;

     // Create fields from fielddefs.
     DestroyFields;
     CreateFields;
end;

// Delete table.
procedure TkbmMemTable.DeleteTable;
begin
     CheckInactive;
     DestroyFields;
end;

procedure TkbmMemTable._InternalAdd(Buffer:Pointer);
begin
     FRecords.Add(Buffer);
end;

procedure TkbmMemTable._InternalInsert(Pos:integer; Buffer:Pointer);
var
   i:integer;
   b:PChar;
begin
     if Pos<0 then Pos:=0;
     if (Pos = FRecords.Count) or(Pos = -1) then
        FRecords.Add(Buffer)
     else
        FRecords.Insert(Pos,Buffer);

     for i:=Pos+1 to FRecords.Count-1 do
     begin
          b:=FRecords.Items[i];
          inc(PRecInfo(b+FStartRecInfo).RecordNo);
     end;
end;

procedure TkbmMemTable._InternalFreeRecord(RecPtr:Pointer);
var
   i:integer;
   pblob:PkbmBlob;
begin
     if FBlobs and (not IsEmpty) then
     begin
          // Browse fields to delete blobs.
          for i:=0 to FieldCount-1 do
          begin
               if Fields[i].DataType in TkbmBlobTypes then
               begin
                    pBlob:=PkbmBlob(GetFieldPointer(RecPtr,Fields[i]));
                    if pBlob^<>nil then
                    begin
                         pBlob^.free;
                         pBlob^:=nil;
                    end;
               end;
          end;
     end;
     FreeMem(RecPtr);
end;

procedure TkbmMemTable._InternalDelete(Pos:integer);
var
   i:integer;
   b:PChar;
begin
     _InternalFreeRecord(FRecords.Items[Pos]);
     FRecords.Delete(Pos);

     for i:=Pos to FRecords.Count-1 do
     begin
          b:=FRecords.Items[i];
          dec(PRecInfo(b+FStartRecInfo)^.RecordNo);
     end;
end;

// Purge all records.
procedure TkbmMemTable._InternalEmpty;
var
   i:integer;
begin
     for i:=0 to FRecords.Count-1 do _InternalFreeRecord(FRecords[i]);
     FRecords.Clear;
end;

procedure TkbmMemTable.InternalOpen;
var
   i,j: integer;
   sz: integer;
begin
     InternalInitFieldDefs;

     // Calculate field offsets into the record.
     FRecordSize:=0;
     FBlobs:=false;                  // Know of no blobs in the definition yet.
     for i:=0 to FieldDefs.Count - 1 do
         with FieldDefs[i] do
         begin
              FFieldOfs[i]:=FRecordSize;

              // Check if a blob.
              if (not FBlobs) and (DataType in TkbmBlobTypes) then FBlobs:=true;

              // Look for fieldsize.
              sz:=CalcFieldSize(DataType,Size);
              inc(sz);   // 1.st byte is boolean flag for Null or not.
              inc(FRecordSize,sz);
         end;

     BindFields(True);
     FRecNo:=-1;
     BookmarkSize:=sizeof(longint);
     FStartRecInfo:=FRecordSize;
     FStartCalculated:=FStartRecInfo+SizeOf(TRecInfo);
     FBufferSize:=FRecordSize+Sizeof(TRecInfo)+CalcFieldsSize;
     FIsOpen:=True;
end;

procedure TkbmMemTable.InternalClose;
begin
     _InternalEmpty;
     FIsOpen:=False;
     BindFields(False);
end;

procedure TkbmMemTable.InternalInitFieldDefs;
var
   i:integer;
begin
     FieldDefs.clear;
     for i:=0 to Fieldcount-1 do
     begin
          FieldDefs.Add(Fields[i].FieldName,Fields[i].DataType,Fields[i].Size,Fields[i].Required);
     end;
end;

function TkbmMemTable.GetActiveRecordBuffer:  PChar;
begin
     case State of
          dsBrowse:        if IsEmpty then
                              Result := nil
                           else
                              Result := ActiveBuffer;
          dsCalcFields:    Result := CalcBuffer;
          dsFilter:        Result:=FFilterBuffer;
          dsEdit,dsInsert: Result:=ActiveBuffer;
     else
          Result:=nil;
     end;
end;

function TkbmMemTable.GetFieldPointer(Buffer:PChar; Field:TField):PChar;
begin
     Result:=Buffer;
     if Buffer=nil then exit;
     if (Field.FieldKind=fkCalculated) or (Field.FieldKind=fkLookup) then
        inc(Result,FStartCalculated+Field.Offset)
     else
        inc(Result,FFieldOfs[Field.FieldNo-1]);
end;

// Result is data in the buffer and a boolean return (true=not null, false=is null).
function TkbmMemTable.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
   SourceBuffer: PChar;
begin
     Result:=False;
     if not FIsOpen then exit;
     SourceBuffer:=GetFieldPointer(GetActiveRecordBuffer,Field);
     if SourceBuffer=nil then Exit;

     if Assigned(Buffer) and (Field.DataType in TkbmSupportedFieldTypes) then
          Move(SourceBuffer[1], Buffer^, CalcFieldSize(Field.DataType,Field.DataSize));

     Result:=boolean(SourceBuffer[0]);
end;

procedure TkbmMemTable.SetFieldData(Field: TField; Buffer: Pointer);
var
   DestinationBuffer: PChar;
begin
     if not FIsOpen then exit;
     DestinationBuffer:=GetFieldPointer(GetActiveRecordBuffer,Field);
     if DestinationBuffer=nil then Exit;

     Boolean(DestinationBuffer[0]):=(Buffer<>nil);

     if Assigned(Buffer) and (Field.DataType in TkbmSupportedFieldTypes) then
        Move(Buffer^,DestinationBuffer[1],CalcFieldSize(Field.DataType,Field.DataSize));

     DataEvent (deFieldChange, Longint(Field));
end;

function TkbmMemTable.BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean;
begin
     Move(BCD^, Curr, SizeOf(Currency));
     Result := True;
end;

function TkbmMemTable.CurrToBCD(const Curr: Currency; BCD: Pointer; Precision, Decimals: Integer): Boolean;
begin
     Move(Curr, BCD^, SizeOf(Currency));
     Result := True;
end;

function TkbmMemTable.IsCursorOpen: Boolean;
begin
     Result:=FIsOpen;
end;

function TkbmMemTable.GetCanModify: Boolean;
begin
     Result:=not FReadOnly;
end;

function TkbmMemTable.GetRecordSize: Word;
begin
     Result:=FRecordSize;
end;

function TkbmMemTable.AllocRecordBuffer: PChar;
begin
     GetMem(Result,FBufferSize);
     FillChar(Result^,FBufferSize,0);
end;

procedure TkbmMemTable.FreeRecordBuffer(var Buffer: PChar);
begin
     FreeMem(Buffer);
end;

procedure TkbmMemTable.InternalFirst;
begin
     _InternalFirst;
end;

procedure TkbmMemTable.InternalLast;
begin
     _InternalLast;
end;

procedure TkbmMemTable._InternalFirst;
begin
     FRecNo:=-1;
end;

procedure TkbmMemTable._InternalLast;
begin
     FRecNo:=FRecords.Count;
end;

function TkbmMemTable._InternalNext:boolean;
begin
     if FrecNo<FRecords.Count-1 then
     begin
          Inc(FRecNo);
          Result:=true;
     end
     else Result:=false;
end;

function TkbmMemTable._InternalPrior:boolean;
begin
     if FrecNo>0 then
     begin
          Dec(FRecNo);
          Result:=true;
     end
     else Result:=false;
end;

function TkbmMemTable.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
   Acceptable: Boolean;
begin
     Result:=grOK;
     Acceptable:=False;
     repeat
         begin
              case GetMode of
                   gmCurrent: begin
                                   if FRecNo>=FRecords.Count then Result:=grEOF
                                   else if FRecNo<0 then Result:=grBOF
                                   else Result:=grOk;
                              end;
                   gmNext:    begin
                                   if _InternalNext then Result:=grOK
                                   else Result:=grEOF;
                              end;
                   gmPrior:   begin
                                   if _InternalPrior then Result:=grOK
                                   else Result:=grBOF;
                              end;
              end;
              if Result=grOk then
              begin
                      //fill TARrecord part of buffer
                      Move(FRecords.Items[FRecNo]^,Buffer^,FBufferSize);

                      //fill information part of buffer
                      with PRecInfo(Buffer+FStartRecInfo)^ do
                      begin
                           RecordNo:=FRecNo;
                           BookmarkFlag:=bfCurrent;
                      end;

                      //fill calc fields part of buffer
                      ClearCalcFields(Buffer);
                      GetCalcFields(Buffer);
                      Acceptable:=FilterRecord(Buffer);
                      if (GetMode=gmCurrent) and not Acceptable then Result:=grError;
              end
         end;
     until (Result<>grOk) or Acceptable;
end;

function TkbmMemTable.FindRecord(Restart, GoForward: Boolean): Boolean;
var
   Status:boolean;
begin
     CheckBrowseMode;
     DoBeforeScroll;
     SetFound(False);
     UpdateCursorPos;
     CursorPosChanged;

     if GoForward then
     begin
          if Restart then _InternalFirst;
          Status := _InternalNext;
     end else
     begin
          if Restart then _InternalLast;
          Status := _InternalPrior;
     end;

     if Status then
     begin
          Resync([rmExact, rmCenter]);
          SetFound(True);
     end;
     Result := Found;
     if Result then DoAfterScroll;
end;

function TkbmMemTable.FilterRecord(Buffer: PChar): Boolean;
var
   SaveState: TDatasetState;
begin
     Result:=True;
     if not Filtered or not Assigned(OnFilterRecord) then Exit;
     SaveState:=SetTempState(dsFilter);
     FFilterBuffer:=Buffer;
     OnFilterRecord(self,Result);
     RestoreState(SaveState);
end;

procedure TkbmMemTable.InternalSetToRecord(Buffer: PChar);
begin
     FRecNo:=PRecInfo(Buffer+FStartRecInfo).RecordNo;
end;

function TkbmMemTable.GetRecordCount: integer;
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not Filtered then Result:=FRecords.Count
     else
     begin
          Result:=0;
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             while GetRecord(TempBuffer,gmNext,True)=grOk do Inc(Result);
          finally
             RestoreState(SaveState);
             FRecNo:=SavePosition;
             FreeRecordBuffer(TempBuffer);
          end;
     end;
end;

function TkbmMemTable.GetRecNo: integer;
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not Filtered then Result:=FRecNo
     else
     begin
          Result:=0;
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             while PRecInfo(TempBuffer+FStartRecInfo).RecordNo<>SavePosition do
               if GetRecord(TempBuffer,gmNext,True)=grOk then Inc(Result);
          finally
             RestoreState(SaveState);
             FRecNo:=SavePosition;
             FreeRecordBuffer(TempBuffer);
          end;
     end;
end;

procedure TkbmMemTable.SetRecNo(Value: Integer);
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not Filtered then FRecNo:=Value
     else
     begin
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             repeat
                   begin
                        if GetRecord(TempBuffer,gmNext,True)=grOk then Dec(Value)
                        else
                        begin
                             FRecNo:=SavePosition;
                             break;
                        end;
                   end;
             until Value=0;
          finally
             RestoreState(SaveState);
             FreeRecordBuffer(TempBuffer);
          end;
     end;

     // refresh the position of the dataset
     Resync([rmCenter,rmExact]);
end;

procedure TkbmMemTable.InternalAddRecord(Buffer: Pointer; Append: Boolean);
var
   b:Pointer;
begin
     // Allocate room for buffer in list.
     GetMem(b,FBufferSize);
     Move(Buffer^, b^, FBufferSize);
     if Append then
        _InternalAdd(b)
     else
         _InternalInsert(FRecNo,b);
end;

procedure TkbmMemTable.InternalDelete;
begin
     _InternalDelete(FRecNo);
end;

procedure TkbmMemTable.InternalInitRecord(Buffer: PChar);
begin
     FillChar(Buffer^,FBufferSize,0);
     PRecInfo(Buffer+FStartRecInfo)^.RecordNo:=FRecNo;
end;

procedure TkbmMemTable.InternalPost;
var
   b:pointer;
   n:integer;
begin
     n:=PRecInfo(ActiveBuffer+FStartRecInfo)^.RecordNo;
     if State = dsEdit then
        Move(ActiveBuffer^, FRecords.Items[n]^, FBufferSize)
     else
     begin
          GetMem(b,FBufferSize);
          Move(ActiveBuffer^, b^, FBufferSize);
          if GetBookmarkFlag(b) = bfEOF then
             _InternalAdd(b)
          else
             _InternalInsert(n,b);
     end;
end;

procedure TkbmMemTable.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
     PRecInfo(Buffer + FStartRecInfo).BookmarkFlag := Value;
end;

function TkbmMemTable.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
     Result:=PRecInfo(Buffer+FStartRecInfo).BookmarkFlag;
end;

procedure TkbmMemTable.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
     PInteger(Data)^ := PRecInfo(Buffer + FStartRecInfo).Bookmark;
end;

procedure TkbmMemTable.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
     PRecInfo(Buffer + FStartRecInfo).Bookmark := PInteger(Data)^;
end;

procedure TkbmMemTable.InternalGotoBookmark (Bookmark: Pointer);
var
  ReqBookmark: Integer;
begin
     ReqBookmark := PInteger (Bookmark)^;
     if (ReqBookmark >= 0) and (ReqBookmark < RecordCount) then
        FRecNo := ReqBookmark
     else
        raise eMemTableError.Create('Bookmark ' + IntToStr(ReqBookmark) + ' not found');
end;

procedure TkbmMemTable.InternalHandleException;
begin
     Application.HandleException(Self);
end;

procedure TkbmMemTable.SaveToFile(const FileName: string; flags:TkbmMemTableSaveFlags);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream,flags);
  finally
    Stream.Free;
  end;
end;

procedure TkbmMemTable.SaveToStream(Stream: TStream; flags:TkbmMemTableSaveFlags);
var
  i:integer;
  bm:TBookmark;
  nf:integer;
  s,a:string;
  l:integer;
  fset,f:^Boolean;
  Ods,Oms:char;
begin
  // Setup standard layout for data.
  Ods:=DateSeparator;
  Oms:=DecimalSeparator;
  DateSeparator:='/';
  DecimalSeparator:='.';

  bm:=GetBookmark;
  fset:=nil;
  try
     DisableControls;

     // Setup flags for fields to save.
     nf:=Fieldcount;
     GetMem(fset,nf * sizeof(boolean));
     f:=fset;
     for i:=0 to nf-1 do
     begin
          f^:=false;
          if not (Fields[i].DataType in TkbmBlobTypes) then
             case Fields[i].FieldKind of
                  fkData: if mtfSaveData in flags then f^:=true;
                  fkCalculated: if mtfSaveCalculated in flags then f^:=true;
                  fkLookup: if mtfSaveLookup in flags then f^:=true;
                  else f^:=true;
             end;
          if not (Fields[i].Visible or (mtfSaveNonVisible in flags)) then f^:=false;
          inc(f);
     end;

     // Write all field display names in CSV format.
     s:='';
     a:='';
     f:=fset;
     for i:=0 to nf-1 do
     begin
          if f^ then
          begin
               s:=s+a+AnsiQuotedStr(PChar(Fields[i].DisplayName),'"');
               a:=',';
          end;
          inc(f);
     end;
     s:=s+#13+#10;
     l:=length(s);
     Stream.Write(Pointer(s)^, l);

     // Write all records in CSV format.
     first;
     while not EOF do
     begin
          // Write current record.
          s:='';
          a:='';
          f:=fset;
          for i:=0 to nf-1 do
          begin
               if f^ then
               begin
                    if (Fields[i].IsNull) then s:=s+a
                    else s:=s+a+AnsiQuotedStr(PChar(Fields[i].AsString),'"');
                    a:=',';
               end;
               inc(f);
          end;
          s:=s+#13+#10;
          l:=length(s);
          Stream.WriteBuffer(Pointer(s)^, l);

          // Next record.
          next;
     end;
  finally
     GotoBookmark(bm);
     EnableControls;
     FreeBookmark(bm);
     if fset<>nil then FreeMem(fset);
     DateSeparator:=Ods;
     DecimalSeparator:=Oms;
  end;
end;

procedure TkbmMemTable.LoadFromFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TkbmMemTable.LoadFromStream(Stream: TStream);
const
   BUFSIZE=8192;
var
   i:integer;
   bm:TBookmark;
   nf:integer;
   s:string;
   buf,bufptr:PChar;
   remaining_in_buf:integer;
   Line:string;
   lptr,elptr:PChar;
   null:boolean;
   Ods,Oms:char;

   function GetChunk:boolean;
   begin
        remaining_in_buf:=Stream.Read(pointer(buf)^,BUFSIZE);
        bufptr:=buf;
        Result:=remaining_in_buf>0;
   end;

   function GetLine:boolean;
   var
     EOL,EOF:boolean;
   begin
        // Cut out a line.
        EOL:=false;
        EOF:=false;
        Line:='';
        while true do
        begin
             // Check if need another chunk.
             if remaining_in_buf=0 then
             begin
                  // Check if EOF.
                  if not GetChunk then
                  begin
                       EOF:=true;
                       break;
                  end;
             end;

             // Check if we got EOL character, skip them and finally break.
             if (bufptr^) in [#0, #10, #13] then EOL:=true
             else if EOL then break;

             // Prepare to look at next char.
             if not EOL then Line:=Line+bufptr^;
             Inc(bufptr);
             dec(remaining_in_buf);
        end;

        lptr:=PChar(Line);
        elptr:=PChar(Line)+Length(Line)-1;
        Result:=(not EOF);
   end;

   function GetWord(var null:boolean):string;
   label
     L_exit;
   begin

     // Cut out next word.
     Result:='';

     // Look for starting " or ,.
     while (lptr^ <> '"') and (lptr^ <> ',') and (lptr<elptr) do inc(lptr);
     if (lptr>=elptr) then exit;
     if (lptr^ = ',') then
     begin
          null:=true;
          inc(lptr);
          exit;
     end
     else null:=false;
     inc(lptr);

     while true do
     begin
          // Look for ending ".
          while not (lptr^ = '"') do
          begin
               if (lptr>=elptr) then goto L_exit;
               Result:=Result+lptr^;
               inc(lptr);
          end;
          inc(lptr);

          // Is it a double "" or end of word ?.
          if (lptr^ = '"') then
          begin
               Result:=Result+'"';
               inc(lptr);
               continue;
          end;

L_exit:
          inc(lptr);
          break;
     end;
   end;

begin
  // Setup standard layout for data.
  Ods:=DateSeparator;
  Oms:=DecimalSeparator;
  DateSeparator:='/';
  DecimalSeparator:='.';

  bm:=GetBookmark;

  try
     // Allocate space for a buffer.
     GetMem(buf,BUFSIZE);

     // Still nothing in the buffer to handle.
     remaining_in_buf:=0;

     // Read data from stream.
     nf:=Fieldcount;

     // Read headerline and skip it.
     GetLine;

     DisableControls;

     // Read all lines in CSV format.
     while true do
     begin
          GetLine;
          if Line='' then break;

          append;

          i:=0;
          while (lptr<elptr) and (i<nf) do
          begin
               s:=GetWord(null);
               if null then Fields[i].Clear
               else Fields[i].AsString:=s;
               inc(i);
          end;

          post;
     end;
  finally
     FreeMem(buf);
     GotoBookmark(bm);
     EnableControls;
     FreeBookmark(bm);
     DateSeparator:=Ods;
     DecimalSeparator:=Oms;
  end;
end;

procedure TkbmMemTable.EmptyTable;
begin
     _InternalEmpty;
end;

procedure TkbmMemTable.SetCommaText(AString: String);
var
   stream:TMemoryStream;
begin
     EmptyTable;
     stream:=TMemoryStream.Create;
     try
        stream.Write(Pointer(AString)^,length(AString));
        stream.Seek(0,soFromBeginning);
        LoadFromStream(stream);
     finally
        stream.free;
     end;
end;

function TkbmMemTable.GetCommaText: String;
var
   stream:TMemoryStream;
   sz:integer;
   p:PChar;
begin
     Result:='';
     stream:=TMemoryStream.Create;
     try
        SaveToStream(stream,[mtfSaveData]);
        stream.Seek(0,soFromBeginning);
        sz:=stream.Size;
        p:=stream.Memory;
        setstring(Result,p,sz);
     finally
        stream.free;
     end;
end;

// Sneak in before the table is closed.
procedure TkbmMemTable.DoBeforeClose;
begin
     inherited;

     // If persistent, save info to file.
     if FPersistent and (FPersistentFile <> '') then
     begin
          try
             SysUtils.DeleteFile(FPersistentFile);
          finally
             SaveToFile(FPersistentFile,[mtfSaveData,mtfSaveNonVisible]);
          end;
     end;
end;

// Sneak in after the table is opened.
procedure TkbmMemTable.DoAfterOpen;
begin
     // If persistent, read info from file.
     if FPersistent and (FPersistentFile <> '') and FileExists(FPersistentFile) then
     begin
          LoadFromFile(FPersistentFile);
          first;
     end;

     inherited;
end;

// Copy records from source to destinations.
// Handles different fieldorder between the two datasets.
// Returns the number of records copied.
function TkbmMemTable.CopyRecords(Source,Destination:TDataSet;Count:longint):longint;
var
   i:integer;
   fc:integer;
   f:TField;
   fi:array [0..255] of integer;
begin
     Result:=0;

     // Did we get valid parameters.
     if (Source=nil) or (Destination=nil) or (Source=Destination) then exit;

     // Build name index relations between destination and source dataset.
     fc:=Destination.FieldCount-1;
     for i:=0 to fc do
     begin
          // Check if not a datafield or not a supported field, dont copy it.
          if (Destination.Fields[i].FieldKind<>fkData) or
             (not (Destination.Fields[i].DataType in TkbmSupportedFieldTypes)) then
          begin
               fi[i]:=-1;
               continue;
          end;

          // Find matching fieldnames on both sides. If fieldname not found, dont copy it.
          f:=Source.FieldByName(Destination.Fields[i].FieldName);
          if f=nil then
          begin
               fi[i]:=-1;
               continue;
          end;

          fi[i]:=f.Index;
     end;

     // Copy data.
     Source.First;
     while not Source.EOF do
     begin
          Destination.Append;
          for i:=0 to fc do
          begin
               if fi[i]>=0 then
                  Destination.Fields[i].Value:=Source.Fields[fi[i]].Value
               else
                  Destination.Fields[i].Clear;
          end;
          Destination.post;
          Source.next;
          inc(Result);
          if (Count>0) and (Result>Count) then break;
     end;
end;

// Fill the memorytable with data from another dataset.
procedure TkbmMemTable.LoadFromDataSet(Source:TDataSet);
var
   SourceActive:boolean;
   SourceDisabled:boolean;
   SelfActive:boolean;
begin
     if Source=self then exit;

     // Close this table.
     SelfActive:=Active;
     Close;

     // Remember state of source.
     SourceActive:=Source.Active;
     SourceDisabled:=Source.ControlsDisabled;

     // Dont update controls while scrolling through source.
     Source.DisableControls;

     try
        DisableControls;
        try

           // Dont want to check filtering while copying.
           Filtered := False;

           // Open source.
           Source.Open;
           Source.CheckBrowseMode;
           Source.UpdateCursorPos;

           // Create this memorytable as a copy of the other one.
           CreateTableAs(Source);
           Open;
           CheckBrowseMode;

           // Move to first record in source.
           Source.First;
           CopyRecords(Source,self,-1);
        finally
           First;
        end;
     finally
        EnableControls;
        if not SourceActive then Source.Close;
        if not SourceDisabled then Source.EnableControls;
        UpdateCursorPos;
        CursorPosChanged;
     end;
end;

// Append the data in this memory table to another dataset.
procedure TkbmMemTable.SaveToDataSet(Destination:TDataSet);
var
   DestActive:boolean;
   DestDisabled:boolean;
   SelfActive:boolean;
begin
     if Destination=self then exit;

     // Close this table.
     SelfActive:=Active;
     Close;

     // Remember state of destination.
     DestActive:=Destination.Active;
     DestDisabled:=Destination.ControlsDisabled;

     // Dont update controls while appending to destination
     Destination.DisableControls;

     try
        DisableControls;
        try
           // Open destination
           Destination.Open;
           Destination.CheckBrowseMode;
           Destination.UpdateCursorPos;

           // Open this if not opened.
           Open;
           CheckBrowseMode;

           // Move to first record in this.
           First;
           CopyRecords(self,Destination,-1);
        finally
           Destination.First;
        end;
     finally
        EnableControls;
        if not DestActive then Destination.Close;
        if not DestDisabled then Destination.EnableControls;
        Active:=SelfActive;
     end;
end;

function TkbmMemTable.IsSequenced: Boolean;
begin
     Result:=not Filtered;
end;

// Record rearranging.

function TkbmMemTable._InternalMoveRecord(Source, Destination: Integer): Boolean;
var
   p:Pointer;
begin
     Result := False;
     if (Source <> Destination) and (Source > -1) and (Source < FRecords.Count)
        and (Destination > -1) and (Destination < FRecords.Count) then
     begin
          p := FRecords[Source];
          FRecords[Source] := FRecords[Destination];
          FRecords[Destination] := p;
          RecNo := Destination;
     end;
end;

// Move record to the specified destination.
function TkbmMemTable.MoveCurRec(Destination: Integer): Boolean;
begin
     Result := _InternalMoveRecord(RecNo,RecNo + Destination);
end;

// Sorting.

// Quick sort algorithm for our purpose.
procedure TkbmMemTable.QuickSort(L, R: Integer);
var
   I, J: Integer;
   P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := FRecords.List^[(L + R) shr 1];
    repeat
      while CompareRecords(FIndexList,FRecords.List^[I], P) < 0 do Inc(I);
      while CompareRecords(FIndexList,FRecords.List^[J], P) > 0 do Dec(J);
      if I <= J then
      begin
        T := FRecords.List^[I];
        FRecords.List^[I] := FRecords.List^[J];
        FRecords.List^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    L := I;
  until I >= R;
end;

// Callback function for TDataset to know if specified field is an index.
function TkbmMemTable.GetIsIndexField(Field:TField):Boolean;
begin
     Result:=FIndexList.IndexOf(Field)>=0;
end;

// Build field list from list of fieldnames.
procedure TkbmMemTable.BuildFieldList(List:TList; const FieldNames: string);
var
   p:integer;
   fld:TField;
begin
     List.Clear;
     p:=1;
     while p<=length(FieldNames) do
     begin
          fld:=FieldByName(ExtractFieldName(FieldNames,p));
          if (fld.FieldKind=fkData) and (fld.DataType in (TkbmSupportedFieldTypes-TkbmBlobTypes)) then
             List.Add(fld)
          else
              DatabaseErrorFmt('Cant index on field %s',[fld.DisplayName]);
     end;
end;

// Do sort on specified indexfields.
procedure TkbmMemTable.Sort;
var
   Pos: TBookmarkStr;
begin
     if (not Active) or (FRecords.Count<=0) then exit;

     BuildFieldList(FIndexList,FIndexFieldNames);
     if FIndexList.Count<=0 then exit;
     
     // Remember our position.
     Pos:=Bookmark;
     try
        QuickSort(0,FRecords.Count-1);
        SetBufListSize(0);
        try
           SetBufListSize(BufferCount + 1);
        except
           SetState(dsInactive);
           CloseCursor;
           raise;
        end;
     finally
        // Reposition.
        Bookmark:=Pos;
     end;
     Resync([]);
end;

// TkbmBlob.

constructor TkbmBlob.Create;
begin
     inherited;
     FLength:=0;
     FBuffer:=nil;
end;

destructor TkbmBlob.Destroy;
begin
     if FBuffer<>nil then FreeMem(FBuffer);
     inherited;
end;

procedure TkbmBlob.LoadFromStream(stream:TMemoryStream);
begin
     FLength:=stream.Size;
     FBuffer:=Allocmem(FLength);
     stream.Seek(0,soFromBeginning);
     stream.Read(FBuffer^,FLength);
end;

procedure TkbmBlob.SaveToStream(stream:TMemoryStream);
begin
     stream.SetSize(FLength);
     stream.Seek(0,soFromBeginning);
     stream.Write(FBuffer^,FLength);
     stream.Seek(0,soFromBeginning);
end;

// TkbmBlobStream

// On create, make a stream access to the specified blobfield in the current record.
constructor TkbmBlobStream.Create(Field:TBlobField;Mode:TBlobStreamMode);
begin
     FField:=Field;
     FFieldNo:=FField.FieldNo;
     FDataSet:=FField.DataSet as TkbmMemTable;
     FMode:=Mode;
     FPBlob:=PkbmBlob(FDataSet.GetFieldPointer(FDataSet.GetActiveRecordBuffer,Field));
     if FPBlob=nil then Exit;
     if Mode<>bmRead then
     begin
          if FField.ReadOnly then DatabaseErrorFmt('Field %s is readonly.',[FField.DisplayName]);
          if not (FDataSet.State in [dsEdit, dsInsert]) then DatabaseError('Dataset is not in edit mode.');
     end;
     if Mode=bmWrite then
        Truncate
     else
        ReadBlobData;
end;

// On destroy, update the blobfield in the current record if the blob has changed.
destructor TkbmBlobStream.Destroy;
begin
     if FModified then
     try
        WriteBlobData;
        FField.Modified:=true;
        FDataSet.DataEvent(deFieldChange,Longint(FField));
     except
        Application.HandleException(Self);
     end;
     inherited Destroy;
end;

procedure TkbmBlobStream.WriteBlobData;
var
   blob:TkbmBlob;
begin
     // Get old allocation if any, and free it.
     if FPBlob^<>nil then
     begin
          FPBlob^.free;
          FPBlob^:=nil;
     end;

     // Make new blob with data and enter it into the field
     blob:=TkbmBlob.Create;
     blob.LoadFromStream(self);
     FPBlob^:=blob;
end;

procedure TkbmBlobStream.ReadBlobData;
var
   blob:TkbmBlob;
begin
     // Get allocation.
     blob:=FPBlob^;
     if blob=nil then exit;

     // Copy the data to the stream.
     blob.SaveToStream(self);
end;

function TkbmBlobStream.Write(const Buffer;Count:Longint): Longint;
begin
     Result:=inherited Write(Buffer,Count);
     if FMode=bmWrite then FModified:=true;
end;

procedure TkbmBlobStream.Truncate;
var
   blob:pointer;
begin
     Clear;

     // If blob allocated, remove allocation.
     if FPBlob^<>nil then
     begin
          FPBlob^.free;
          FPBlob^:=nil;
     end;

     FModified:=true;
end;


procedure Register;
begin
     RegisterComponents('Data Access', [TkbmMemTable]);
end;

end.
