unit mcComboBox;

{ =======================================================================================

  Copyright 1997 - 1998, eMCee Software
  mrichter@emceesoftware.com

  Internal development, Delphi 3, Delphi 4 (not tested under previous versions)

    +--------- Internal development
    | +------- Initial release
    | | +----- Production
    | | |  +-- Build
    v v v  v
  v 0.1.3.86 (Internal development, released for use, available for distribution)

  =======================================================================================

  Purpose:

    This unit defines a TComboBox descendant that provides the ability to search for an item within the list of items
    in the combo box.  Partial matches are performed locating the first item in the list which matches its first
    'Length(SearchKey)' characters.  This search is performed from index 0 and progresses forward.  Therefore, a sorted
    list will return the first 'natural' match, while an unsorted will return the first one located which satisfies the
    criteria.

  Function:

    All of the standard TComboBox has been retained and the following extensions have been added:

      A new event 'OnAddNewItem' has been added to notify when the search text doesn't match any existing item.  The new item
      will not be automatically added to the list.  It is the responsibility of the calling program to add the item.

      Note that when the search text no longer matches the items in the list, the original character case is restored.  This
        way, searching for items which you *want* to have a different case will work.  i.e. searching for 'deLuca' when the
        list already contains 'Derickson', at first, 'De' will be highlighted with 'Derickson' showing in the list.  However,
        when the 'L' is typed, the text 'deL' will be displayed.
      If the name 'Derickson' is displayed, and the mouse or arrow keys are used to highlight 'De', then the letter 'L' is
        typed, the text displayed (and the search text) will be 'DeL'.  This is due to the necessity to handle the user
        randomly manipulating the displayed text.  At this point, there is no way to know that a different case 'd' was
        expected by the user.

      The [Escape] key erases the current text and resets the search (start over).
      The [Insert] key will toggle between 'insert' and 'overstrike' modes.  No indicator is currently given to indicate
        the current mode for the control (cannot find how to change blinking cursor).  Note however, that the 'insert' mode
        is reset on entering the control, therefore it is fairly predictable to the user.
      The [Backspace] key will, in a special case, remove one letter from the search text and initiate a new search.  In one
        case (text highlighted from the beginning), [Backspace] will prune at the end of the highlighted text.
      On entering the component, the search text is reset and a new search is initiated.
      Highlighting the entire text beyond any existing key will reset the search text and initiate a new search.
      Pressing the [Enter] key simulates the 'DoExit' event (look at calling DoExit instead).  If a match is not found, the
        OnAddNewItem event is called.
      On exiting the component, if a match is not found, the OnAddNewItem event is called.

  Known Issues:

    Currently, the text typed is highlighted and the 'completion' text is not.  This is opposite to the manner in which
    Quicken displays their search, and seems somewhat counter-intuitive.

  =======================================================================================

    This component is offered as "freeware" and is released into the public domain with the
    exception that no copyright may be sought by anyone other than eMCee Software.

    You may use this component and its source code in any manner you wish with the
    understanding that it is provided "as is" without warranty of any kind.  In no manner,
    and under no circumstances, will eMCee Software be held liable for any damages which
    may arise from the use of this component.

    If you have any problems with the component, or have any suggestions for improvement, please
    contact eMCee Software (mrichter@emceesoftware.com).

    Particular thanks to the members of TeamB, without whom, the knowledge of these techniques
    would be extremely difficult to find.

  =======================================================================================

    Particular thanks to the members of TeamB, without whom, the knowledge of these techniques
    would be extremely difficult to find.

  ======================================================================================= }

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls;

type
  { -- event type declaraction }
  {    even though this version doesn't automatically add item to list, keep 'var' parameter for backward compatability }
  TNewItemEvent = procedure(Sender: TObject; var NewItem: String) of object;
  TmcComboBox = class(TComboBox)
  private
    { Private declarations }
    FOnAddNewItem : TNewItemEvent;
  protected
    { Protected declarations }
    { -- override these to provide enhanced functionality }
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure DoExit; override;
    procedure DoEnter; override;
    procedure Change; override;
  public
    { Public declarations }
  published
    { Published declarations }
    { -- add an event to make 'not found' notifications easier }
    property OnAddNewItem: TNewItemEvent read FOnAddNewItem write FOnAddNewItem;
  end;

procedure Register;

implementation

{ == register the component == }

{ ============================================================================= }
procedure Register;
begin
  RegisterComponents('eMCee', [TmcComboBox]);
end;

{ == include files and declare global variables == }

const
  InsertMode : Boolean = True;
  SearchText : String = '';

{ == internal procedure handlers == }

{ ============================================================================= }
procedure ProcessComboSearch(Sender: TObject; SearchString: String; var Key: Char);
var
  i : integer;
  SavedSelStart : Integer;
begin

  { -----
    This routine performs the actual lookup.
    If a match is found, the first Length(SearchKey) characters are highlighted, otherwise no highlight is shown
    ----- }

  with TmcComboBox(Sender)
  do
  begin

    { -- can't work with an empty list }
    if Items.Count = 0
    then
      Exit;

    { -- make sure the search string has something in it }
    if Length(SearchString) > 0
    then
    begin
      { step through the list, in order }
      for i := 0 to Items.Count - 1
      do
      begin
        { see if this item begins with the SearchString }
        if UpperCase(Copy(Items[i], 1, Length(SearchString))) = UpperCase(SearchString)
        then
        begin
          { point to the match }
          ItemIndex := i;
          { display it }
          Text := Items[ItemIndex];
          { consume the keystroke }
          Key := chr(0);
          { highlight the matching portion of the item }
          SelStart := 0;
          SelLength := Length(SearchString);
          { were're done }
          Exit;
        end;
      end;
    end;

    { -- if the search string is empty, or the search string wasn't found }
    { note the cursor position }
    SavedSelStart := SelStart;
    { point to 'nothing' }
    ItemIndex := -1;
    { display the current search string }
    Text := SearchString;
    { put the cursor back }
    SelStart := SavedSelStart;
    { consume the keystroke }
    Key := chr(0);

  end;

end;

{ == methods == }

{ == event handlers == }

{ ============================================================================= }
procedure TmcComboBox.Change;
begin

  { -- reset the search text }
  SearchText := '';

  Inherited Change;

end;

{ ============================================================================= }
procedure TmcComboBox.DoExit;
var
  NewItem : string;
begin

  { -- if the search text is not found, and the displayed text is not empty, alert the program }
  if (ItemIndex = -1) and
     (Trim(SearchText) <> '')
  then
  begin
    NewItem := Text;
    if Assigned(FOnAddNewItem)
    then
      FOnAddNewItem(Self, NewItem);
  end;

  { -- reset the search text }
  SearchText := '';

  Inherited DoExit;

end;

{ ============================================================================= }
procedure TmcComboBox.DoEnter;
begin

  { -- reset the search text }
  SearchText := '';
  { -- make sure insert mode is on (standard Windows behavior) }
  InsertMode := True;

  Inherited DoEnter;

end;

{ ============================================================================= }
procedure TmcComboBox.KeyDown(var Key: Word; Shift: TShiftState);
var
  KeyChar : char;
  NewItem : String;
  SavedSelStart : integer;
begin


  { capture the keystroke to pass to the search routine }
  KeyChar := chr(Key);

  { override certain keystrokes }
  case Key
  of
    { enter key }
    VK_Return :
      begin
        { -- make one last attempt to find a match }
        { capture the displayed text (so the component uses what the user sees) }
        SearchText := Text;
        { perform the search }
        ProcessComboSearch(Self, SearchText, KeyChar);
        { place the cursor at the end of the displayed text }
        SelLength := 0;
        SelStart := GetTextLen;
        { -- if the search text is not found, and the displayed text is not empty, alert the program }
        if (ItemIndex = -1) and
           (Trim(SearchText) <> '')
        then
        begin
          NewItem := Text;
          if Assigned(FOnAddNewItem)
          then
            FOnAddNewItem(Self, NewItem);
          { reset the search text }
          SearchText := '';
        end;
        { toggle the enabled state to avoid cursor problems (check this to make sure we still need to do this...) }
        Enabled := False;
        Enabled := True;
        { set the focus back to the control (it seems to wander sometimes) }
        SetFocus;
        { consume the [Enter] key to stop the 'ding' (it will no longer be visible to the user's OnKeyDown event) }
        Key := 0;
      end;
    { backspace key }
    VK_Back :
      begin
        { save the cursor position }
        SavedSelStart := SelStart;
        { -- if the beginning of the text is highlighted, prune the search text back to the highlighted portion only }
        if (SelStart = 0) and
           (SelLength > 0)
        then
        begin
          { set the search text to the highlighted portion of the displayed text }
          SearchText := Copy(Text, 1, SelLength-1);
          { position the cursor at the end of the text }
          SelStart := Length(SearchText);
        end
        else
        { -- otherwise, simulate normal behavior }
        begin
          { if any portion is highlighted }
          if SelLength > 0
          then
          begin
            { remove it from the display }
            Text := Copy(Text, 1, SelStart) + Copy(Text, SelStart+SelLength+1, GetTextLen);
            { reposition the cursor }
            SelStart := SavedSelStart;
          end
          else
          { otherwise }
          begin
            { remove the character just to the left of the previous cursor }
            Text := Copy(Text, 1, SelStart-1) + Copy(Text, SelStart+1, GetTextLen);
            { reposition the cursor one position to the left }
            if SavedSelStart > 0
            then
              SelStart := SavedSelStart - 1;
          end;
          { update the search text since we've done random editing }
          SearchText := Text;
        end;
        { see if the new search text is found }
        ProcessComboSearch(Self, SearchText, KeyChar);
        { consume the backspace keystroke (can't pass it on as we've already done what it is supposed to do) }
        Key := 0;
      end;
    { escape key }
    VK_Escape :
      begin
        { -- reset the search text, clear the displayed text, and point to 'nothing' (start over) }
        SearchText := '';
        Text := '';
        ItemIndex := -1;
      end;
    { insert key }
    VK_Insert :
      begin
        { toggle the effective insert mode }
        InsertMode := not InsertMode;
      end;
    { delete key }
    VK_Delete :
      begin
        { save the cursor position }
        SavedSelStart := SelStart;
        { -- if nothing is highlighted, duplicate normal behavior: delete single character to right, or prune to end }
        if SelLength = 0
        then
        begin
          { if the control key is held down }
          if Shift = [ssCtrl]
          then
            { prune }
            SearchText := Copy(Text, 1, SelStart)
          else
            { otherwise, delete single character to the right }
            SearchText := Copy(Text, 1, SelStart) + Copy(Text, SelStart+SelLength+2, GetTextLen);
        end
        else
        { otherwise, something is highlighted }
        begin
          { keep everything except what is highlighted }
          SearchText := Copy(Text, 1, SelStart) + Copy(Text, SelStart+SelLength+1, GetTextLen);
        end;
        SelStart := SavedSelStart;
        { consume the delete keystroke (can't pass it on as we've already done what it is supposed to do) }
        Key := 0;
        { perform the search }
        ProcessComboSearch(Self, SearchText, KeyChar);
      end;
    { up arrow key }
    VK_Up:
      begin
        { -- moving to the previous item in the list }
        { reset the search text }
        SearchText := '';
      end;
    { down arrow key }
    VK_Down:
      begin
        { -- moving to the next item in the list }
        { reset the search text }
        SearchText := '';
      end;
  end;

  Inherited KeyDown(Key, Shift);

end;

{ ============================================================================= }
procedure TmcComboBox.KeyPress(var Key: Char);
var
  SearchKey : string[1];
  SavedSelStart : Integer;
begin

  { if we have an [Enter] key, or the key has been consumed, bail out }
  if (Key = chr(13)) or
     (Key = chr(0))
  then
  begin
    { reset the search text (a precaution due to some earlier side effects) }
    SearchText := '';
    Inherited KeyPress(Key);
    Exit;
  end;

  SearchKey := '';

  { -- if the keystroke is a 'normal' typed key }
  if (ord(Key) > $1f) and (ord(key) < $7f)
  then
  begin
    { capture the keystroke }
    SearchKey := Key;
    { --  if all text is highlighted, but the search text is different, assume this is a new search }
    if (SelStart = 0) and
       (SelLength = Length(SelText)) and
       (GetTextLen > 0)
    then
    begin
      if (UpperCase(SearchText) <> UpperCase(Text))
      then
        { the new search builds on the old search }
        SearchText := SearchText + SearchKey
      else
        { otherwise, the new search text becomes the key pressed }
        SearchText := SearchKey;
      { move the cursor to the end }
      SelStart := Length(SearchText);
    end
    else
    { -- otherwise, if text is highlighted beginning with the first character, use that as the basis for the search text }
    {    i.e. add the key pressed to the highlighted text and continue the search pattern }
    if (SelStart = 0) and
       (SelLength > 0)
    then
    begin
      { make sure the search text agrees with the highlighted text (in case of random highlighting and editing) }
      SearchText := Copy(Text, 1, SelLength) + SearchKey;
      { if the search text is longer than the displayed text, display the search key instead }
      if Length(SearchText) > GetTextLen
      then
        Text := SearchText;
      { position the cursor at the end of the search text (the highlighted portion) }
      SelStart := Length(SearchText);
    end
    else
    { -- otherwise, assume a random edit is occurring }
    begin
      { Save the cursor position }
      SavedSelStart := SelStart;
      { if none of the text is highlighted }
      if SelLength = 0
      then
      begin
        { use different techniques depending on whether insert mode is active or not }
        if InsertMode
        then
          { insert the character and stick the rest of the text back on }
          SearchText := Copy(Text, 1, SelStart) + SearchKey + Copy(Text, SelStart+1, GetTextLen)
        else
          { insert the character, jump over one character, and stick the rest of the text back on }
          SearchText := Copy(Text, 1, SelStart) + SearchKey + Copy(Text, SelStart+2, GetTextLen)
      end
      else
      { otherwise, some text is highlighted }
      begin
        { remove the highlighted portion of the text, and insert the character in its place }
        SearchText := Copy(Text, 1, SelStart) + SearchKey + Copy(Text, SelStart+SelLength, GetTextLen);
      end;
      { update the displayed text }
      Text := SearchText;
      { reposition the cursor to past the character just typed }
      SelStart := SavedSelStart + 1;
      SelLength := 0;
    end;

  end;

  { perform the search }
  ProcessComboSearch(Self, SearchText, Key);

  Inherited KeyPress(Key);

end;

end.

