unit mcListBox;

{ =======================================================================================

  eMCee Software
  mrichter@emceesoftware.com

  Internal development, Delphi 3, Delphi 4 (not tested under previous versions)

    +--------- Internal development
    | +------- Initial release
    | | +----- Beta
    | | |  +-- Build
    v v v  v
  v 0.1.2.27 (Internal development, beta test phase, available for distribution)

  =======================================================================================

  Purpose:

    This unit defines a TListBox descendant that provides two missing features in the standard
    VCL TListBox component, a horizontal scrollbar, and a popup window that displays text that
    is wider than the window.

  Function:

    This listbox component automatically detects whether any item is longer than the width of
    the listbox on the form, and creates a horizontal scroll bar.  Additionally, when the mouse
    cursor passes over any item that is too long to be displayed, a popup hint is displayed
    that contains the entire text of the item.

  Known Issues:

    This component currently does not:
    - prevent the popup window from scrolling off the right side of the screen
    - display the popup window for shorter items when the scrollbar is moved toward the right
    - display the popup window if an item is shorter than the width of the window, but the end
      is partially hidden by the vertical scroll bar
    - the border around the popup window doesn't always display for the area outside the listbox
      component

  =======================================================================================

    This component is offered as "freeware" and is released into the public domain with the exception
    than no copyright may be sought except by the original authors.

    You may use this component and its source code in any manner you wish with the
    understanding that it is provided "as is" without warranty of any kind.  In no manner,
    and under no circumstances, will eMCee Software be held liable for any damages which
    may arise from the use of this component.

    If you have any problems with the component, or have any suggestions for improvement, please
    contact eMCee Software (mrichter@emceesoftware.com).

  =======================================================================================

    Special thanks are extended to Gregor Smith for the implementation of the horizontal scroll bar.
    His extremely clever coding overcomes significant problems encountered when trying to override
    the Items methods.

    Particular thanks to the members of TeamB, without whom, the knowledge of these techniques
    would be extremely difficult to find.

  ======================================================================================= }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TmcListBox = class(TListBox)
  { horizontal scrollbar stuff }
  { -----
    This portion of the component is derived directly from Gregor Smith's
    work in ThsListBox (gregor_smith@bigfoot.com).
    His code was made available as "freeware" and distributed on the Delphi Super Page.
    Although he didn't specifically claim a copyright, he deserves appropriate recognition.
    ----- }
  private
    MaxExtent: integer;
    function GetStringWidth( s: string ): integer;
    procedure LBAddString( var Message: TMessage ); message LB_ADDSTRING;
    procedure LBInsertString( var Message: TMessage ); message LB_INSERTSTRING;
    procedure LBDeleteString( var Message: TMessage ); message LB_DELETESTRING;
    procedure LBResetContent( var Message: TMessage ); message LB_RESETCONTENT;
  { popup text stuff }
  private
    PopupText : THintWindow;
    procedure WMMouseLeave(Var Msg:TWmMouse); Message CM_MouseLeave;
  public
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

procedure Register;

implementation

{ ============================================================================= }
procedure Register;
begin
  RegisterComponents('eMCee', [TmcListBox]);
end;

{ ============================================================================= }
{ ============================================================================= }
{ ===== horizontal scroll bar stuff ===== }
{ ============================================================================= }
{ ============================================================================= }

{ ============================================================================= }
function TmcListBox.GetStringWidth( s: string ): integer;
var
  hDCTemp: HDC;
  hFontOld, hFontTemp: HFONT;
  tm: TTextMetric;
  size: TSize;
begin
  // get the temporary dc for the list box
  hDCTemp := GetDC( Handle );
  // Get a handle to the font used in the list box
  hFontTemp := SendMessage( Handle, WM_GETFONT, 0, 0 );
  // Select the list box font into the temporary DC
  hFontOld := SelectObject( hDCTemp, hFontTemp );

  GetTextMetrics( hDCTemp, tm );
  GetTextExtentPoint32( hDCTemp, PChar(s), Length(s), size );

  // Restore Original Font
  SelectObject( hDCTemp, hFontOld );
  ReleaseDC( Handle, hDCTemp );
  // calculate the width
  Result := size.cx + tm.tmAveCharWidth;
end;

{ ============================================================================= }
procedure TmcListBox.LBAddString( var Message: TMessage );
var
  StrExtent: integer;
begin
  StrExtent := GetStringWidth( PChar(Message.lParam) );
  if StrExtent > MaxExtent then
  begin
    MaxExtent := StrExtent;
    SendMessage( Handle, LB_SETHORIZONTALEXTENT, MaxExtent, 0 );
  end;
  inherited;
end;

{ ============================================================================= }
procedure TmcListBox.LBInsertString( var Message: TMessage );
var
  StrExtent: integer;
begin
  StrExtent := GetStringWidth( PChar(Message.lParam) );
  if StrExtent > MaxExtent then
  begin
    MaxExtent := StrExtent;
    SendMessage( Handle, LB_SETHORIZONTALEXTENT, MaxExtent, 0 );
  end;
  inherited;
end;

{ ============================================================================= }
procedure TmcListBox.LBDeleteString( var Message: TMessage );
var
  StrExtent: integer;
  Index: Integer;
  i : Integer;
begin
  Index := Message.wParam;
  StrExtent := GetStringWidth( PChar(Items[ Index ]) );
  if StrExtent = MaxExtent then
  begin
    MaxExtent := 0;
    for i:= 0 to Items.Count - 1 do begin
      if i <> Index then begin
          StrExtent := GetStringWidth( PChar(Items[ i ]) );
          if StrExtent > MaxExtent then begin
            MaxExtent := StrExtent;
          end;
      end;
    end;
    SendMessage( Handle, LB_SETHORIZONTALEXTENT, MaxExtent, 0 );
  end;
  inherited;
end;

{ ============================================================================= }
procedure TmcListBox.LBResetContent( var Message: TMessage );
begin
    MaxExtent := 0;
    SendMessage( Handle, LB_SETHORIZONTALEXTENT, MaxExtent, 0 );
    inherited;
end;

{ ============================================================================= }
{ ============================================================================= }
{ ===== hint popup stuff ===== }
{ ============================================================================= }
{ ============================================================================= }

{ ============================================================================= }
constructor TmcListBox.Create(AOwner: TComponent);
begin

  { create the component }
  Inherited Create(AOwner);

  { create the popup window }
  PopupText := THintWindow.Create(Self);
  { must assign the popup window's parent as the listbox, or it doesn't display }
  PopupText.Parent := Self;

end;

{ ============================================================================= }
destructor TmcListBox.Destroy;
begin

  { free the popup window }
  PopupText.Free;

  { free the listbox }
  inherited Destroy;

end;

{ ============================================================================= }
procedure TmcListBox.WMMouseLeave(Var Msg:TWmMouse);
begin

  { hide the popup window }
  PopupText.Hide;

  Inherited;

end;

{ ============================================================================= }
procedure TmcListBox.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Point : TPoint;
  MouseOverItem : Integer;
begin

  { call the inherited event first }
  Inherited;

  { determine which item in the list the mouse is over }
  MouseOverItem := TopIndex + (y div ItemHeight);

  { if the mouse is past the end of the list, bail out }
  if MouseOverItem > Items.Count-1
  then
  begin
    PopupText.Hide;
    Exit;
  end;

  { if the string is not wider than the listbox, bail out }
  if GetStringWidth(Items[MouseOverItem]) < Width
  then
  begin
    PopupText.Hide;
    Exit;
  end;

  { write the text onto the popup window so that it doesn't appear to "move" }
  PopupText.Canvas.TextOut(1, -1, Items[MouseOverItem]);

  { the coordinates for the popup window are relative to the screen, so we need to convert from component to screen }
  Point.x := 0;
  Point.y := (y div ItemHeight) * ItemHeight;
  Point := ClientToScreen(Point);

  { position the popup window over the long item in the listbox }
  PopupText.Left := Point.x;
  PopupText.Top := Point.y;
  { make it slightly wider than the text to allow space before and after }
  PopupText.Width := GetStringWidth(Items[MouseOverItem]) + 2;
  { make it slighter higher than the listbox item height so it appears more balanced on the screen }
  PopupText.Height := ItemHeight + 1;

  { display the popup window }
  PopupText.Show;

end;

end.

